"""
Display a random array of colors for 5 seconds.
"""

from lpminimk3 import Mode, find_launchpads
import random
import time


def main():
    lp = find_launchpads()[0]  # Get the first available launchpad
    lp.open()  # Open device for reading and writing on MIDI interface (by default)  # noqa

    lp.mode = Mode.PROG  # Switch to the programmer mode

    for led in lp.panel.led_range():  # Loop through all LEDs
        led.color = random.randint(1, 127)  # Set LED to a random color  # noqa

    time.sleep(5)  # Keep LEDs on for a while

    for led in lp.panel.led_range():
        del led.color  # Turn off LED


if __name__ == '__main__':
    main()



import os

_CURRENT_DIR = os.path.dirname(__file__)


class Bitmaps:
    SMILEY = os.path.join(_CURRENT_DIR, '../bitmaps/smiley.bitmap.json')  # noqa


class Movies:
    PING_PONG = os.path.join(_CURRENT_DIR, '../movies/ping_pong.movie.json')  # noqa



"""
Graphics module for high-level rendering on the Launchpad Mini MK3's surface.
"""

from ._graphics import Text,\
                       Bitmap,\
                       Movie,\
                       Frame,\
                       Renderable,\
                       ScrollDirection,\
                       FlipAxis  # noqa



"""
Region of LEDs on a matrix.
"""

from abc import ABC


class Region(ABC):
    """
    A collection of LEDs on a matrix.
    """

    def __len__(self):
        return len(self.button_names)

    @property
    def button_names(self):
        """
        Button names for region.
        """
        pass


class Labeled(Region):
    """
    All labeled buttons on a panel.
    """

    @Region.button_names.getter
    def button_names(self):
        return ['up',
                'down',
                'left',
                'right',
                'session',
                'drums',
                'keys',
                'user',
                'logo',
                'scene_launch_1',
                'scene_launch_2',
                'scene_launch_3',
                'scene_launch_4',
                'scene_launch_5',
                'scene_launch_6',
                'scene_launch_7',
                'stop_solo_mute']



"""
Sync server for controlling connected Launchpad Mini MK3 clients.
"""

import asyncio
import websockets
import json
import sys
from collections import namedtuple
from argparse import ArgumentParser
from functools import partial
from lpminimk3 import find_launchpads, Mode
from lpminimk3.graphics import Frame


_connected_clients = []


async def handler(lps, websocket, path):
    if path == "/sync":
        _connected_clients.append(websocket)
        try:
            while True:
                data = await websocket.recv()
                json_data = json.loads(data)
                for lp in lps:
                    lp.grid.render(Frame(json_data))
                websockets.broadcast(_connected_clients, data)
        finally:
            _connected_clients.remove(websocket)


async def sync_with_sketch(lps, host, port):
    async with websockets.serve(partial(handler, lps), host, port):
        await asyncio.Future()


def find_lps():
    lps = []
    for lp in find_launchpads():
        lp.open()
        lp.mode = Mode.PROG
        lps.append(lp)
    return lps


def init_parser(ip, port):
    parser = ArgumentParser(description="Sync server for lpminimk3")
    parser.add_argument("-a",
                        "--all",
                        action="store_true",
                        help="Listen for all hosts (on 0.0.0.0)")
    parser.add_argument("-i",
                        "--ip",
                        help=f"IP to serve on (Default: {ip})")
    parser.add_argument("-p",
                        "--port",
                        type=int,
                        help=f"Port to bind to (Default: {port})")
    parser.add_argument("--version",
                        action="version",
                        version="%(prog) 0.1")
    return parser


async def main(*, ip="localhost", port=7654, allow_all_hosts=False):
    Args = namedtuple("Args", ["ip", "port", "all"])
    args = Args(ip, port, allow_all_hosts)
    parser = init_parser(ip, port)
    try:
        if len(sys.argv) > 1:
            args = parser.parse_args(sys.argv[1:])
        lps = find_lps()
        ip = "0.0.0.0" if args.all else args.ip
        ip = args.ip if args.ip else ip
        port = args.port if args.port else port
        if args.ip and args.all:
            print("Can't set both '--ip' and '--all'.", file=sys.stderr)
            parser.print_help()
            return 1
        else:
            await sync_with_sketch(lps, ip, port)
    except KeyboardInterrupt:
        return 1


if __name__ == "__main__":
    asyncio.run(main())



"""
Wait for button presses and log events.
"""

from lpminimk3 import ButtonEvent, Mode, find_launchpads
import random
import sys


def handle_event(button_event):
    if button_event and button_event.type == ButtonEvent.PRESS:
        button_event.button.led.color = random.randint(1, 127)  # Set LED to random color while button is pressed  # noqa
        print(f"Button '{button_event.button.name}' pressed.")
    elif button_event and button_event.type == ButtonEvent.RELEASE:
        button_event.button.led.color = 0  # Turn LED off once button is released  # noqa
        print(f"Button '{button_event.button.name}' released.")
    else:
        sys.exit()  # Exit on KeyboardInterrupt


def main():
    lp = find_launchpads()[0]  # Get the first available launchpad
    lp.open()  # Open device for reading and writing on MIDI interface (by default)  # noqa

    lp.mode = Mode.PROG  # Switch to the programmer mode

    print('Push any button on your Launchpad to see it light up!\n'
          'Press Ctrl+C to quit.\n')
    while True:
        handle_event(lp.panel.buttons().poll_for_event())  # Wait for a button press/release  # noqa


if __name__ == '__main__':
    main()



"""
Software representation of physical components for Launchpad Mini MK3.
"""

import math
import re
from abc import ABC
from ..colors._colors import ColorShade, ColorShadeStore, RgbColor
from ..midi_messages import Colorspec,\
                            ColorspecFragment,\
                            Constants,\
                            Lighting
from ..match import ButtonMatch
from .utils import ButtonEvent
from ..region import Region


class Matrix(ABC):
    """
    A matrix of pad buttons/LEDs on the surface of the Launchpad.
    """
    @property
    def launchpad(self):
        pass

    @property
    def width(self):
        return -1

    @property
    def height(self):
        return -1

    @property
    def max_x(self):
        return -1

    @property
    def max_y(self):
        return -1

    def led_range(self):
        pass

    def render(self, renderable):
        renderable.render(self)


class FlipAxis:
    X = 'x'
    Y = 'y'
    XY = 'xy'


class _MatrixTransform:
    def __init__(self, matrix, layout, led_mode):
        self._matrix = matrix
        self._layout = layout
        self._led_mode = led_mode

    def rotated_led_range(self, angle, *, flip_axis=''):
        assert angle, 'Angle cannot be zero.'
        angle = self._normalize_angle(angle)
        angle = self._flip_angle(angle)
        angle_rad = math.radians(angle)
        for led in self._matrix.led_range(layout=self._layout,
                                          mode=self._led_mode):
            rotated_x = round((led.y * math.sin(angle_rad)) + (led.x * math.cos(angle_rad)))  # noqa
            rotated_y = round((led.y * math.cos(angle_rad)) - (led.x * math.sin(angle_rad)))  # noqa
            if angle == 90:
                rotated_y = min(self._matrix.max_y + rotated_y, self._matrix.max_y)  # noqa
            elif angle == 180:
                rotated_y = min(self._matrix.max_y + rotated_y, self._matrix.max_y)  # noqa
                rotated_x = min(self._matrix.max_x + rotated_x, self._matrix.max_x)  # noqa
            elif angle == 270:
                rotated_x = min(self._matrix.max_x + rotated_x, self._matrix.max_x)  # noqa
            if flip_axis:
                flipped_x = (self._flip_axis(rotated_x)
                             if FlipAxis.X in flip_axis
                             else rotated_x)
                flipped_y = (self._flip_axis(rotated_y)
                             if FlipAxis.Y in flip_axis
                             else rotated_y)
                yield self._matrix.led(x=flipped_x,
                                       y=flipped_y,
                                       layout=self._layout,
                                       mode=self._led_mode)
            else:
                yield self._matrix.led(x=rotated_x,
                                       y=rotated_y,
                                       layout=self._layout,
                                       mode=self._led_mode)

    def flipped_led_range(self, axis):
        for led in self._matrix.led_range(layout=self._layout,
                                          mode=self._led_mode):
            flipped_x = (self._flip_axis(led.x)
                         if FlipAxis.X in axis
                         else led.x)
            flipped_y = (self._flip_axis(led.y)
                         if FlipAxis.Y in axis
                         else led.y)
            yield self._matrix.led(x=flipped_x,
                                   y=flipped_y,
                                   layout=self._layout,
                                   mode=self._led_mode)

    def _flip_axis(self, value):
        max_value = self._matrix.width - 1
        return max_value - value

    def _flip_angle(self, angle):
        if angle == -90 or angle == 270:
            return 90
        elif angle == -180:
            return 180
        elif angle == 90 or angle == -270:
            return 270
        return angle

    def _normalize_angle(self, angle):
        if angle and (angle % 90) == 0 and abs(angle) > 270:
            return angle % 360
        assert (round(abs(angle)) in (0, 90, 180, 270)), \
               'Angle must be a multiple of 90.'
        return angle


class _MatrixCoordinate:
    def __init__(self, launchpad, layout, button_names, *,
                 name='',
                 coordinate_id=-1,
                 x=-1, y=-1):
        x, y = (self._determine_coordinate_from_name(name,
                                                     button_names)
                if x < 0 and y < 0 and len(name) > 0
                else (x, y))
        matrix_width, matrix_height = self._determine_bounds(button_names)

        x, y = (self._determine_coordinate_from_id(coordinate_id,
                                                   bounds=(matrix_width, matrix_height))  # noqa
                if coordinate_id >= 0 and x < 0 and y < 0
                else (x, y))
        name = self._determine_name_from_coordinate(x, y,
                                                    button_names,
                                                    bounds=(matrix_width, matrix_height))  # noqa
        midi_value = self._determine_midi_value(x, y,
                                                layout,
                                                bounds=(matrix_width, matrix_height))  # noqa
        self._x = x
        self._y = y
        self._id = coordinate_id
        self._matrix_width = matrix_width
        self._matrix_height = matrix_height
        self._name = name
        self._midi_value = midi_value

    def __repr__(self):
        return ('LayoutCoordinate('
                f'name={self.name}, '
                f'x={self.x}, '
                f'y={self.y}, '
                f'id={self.id})')

    @property
    def id(self):
        within_range = (self._x >= 0
                        and self._y >= 0
                        and self._x < self._matrix_width
                        and self._y < self._matrix_height)
        if not within_range:
            return -1
        return (self._y * self._matrix_height) + self._x + 1

    @property
    def name(self):
        return self._name

    @property
    def x(self):
        return self._x

    @property
    def y(self):
        return self._y

    @property
    def midi_value(self):
        return self._midi_value

    @property
    def matrix_width(self):
        return self._matrix_width

    @property
    def matrix_height(self):
        return self._matrix_height

    def _determine_coordinate_from_id(self, led_id, bounds):
        width, height = bounds
        x = int(led_id % width)
        y = int(led_id / height)
        return x, y

    def _determine_coordinate_from_name(self, name, button_names):
        found = False
        x = y = -1
        for column, button_column in enumerate(button_names):
            for row, button_name in enumerate(button_column):
                if button_name == name.lower():
                    x = row
                    y = column
                    found = True
                    break
                elif (name.lower() in button_name
                        and name.lower()
                        in ButtonFace.STOP_SOLO_MUTE.split('_')):
                    x = row
                    y = column
                    found = True
                    break
            if found:
                break
        if x < 0 or y < 0:
            raise ValueError('Invalid name set.')

        return x, y

    def _determine_midi_value(self, x, y, layout, bounds):
        width, height = bounds
        within_range = (x >= 0
                        and y >= 0
                        and x < width
                        and y < height)
        if not within_range:
            raise ValueError('Led(x,y) out of range: '
                             f'value({x},{y}), '
                             f'range((0,{width}),(0,{height}))')
        return layout[y][x]

    def _determine_name_from_coordinate(self, x, y, button_names, bounds):
        matrix_width, matrix_height = bounds
        within_range = (x >= 0
                        and y >= 0
                        and x < matrix_width
                        and y < matrix_height)
        if not within_range:
            raise ValueError('Led(x,y) out of range: '
                             f'value({x},{y}), '
                             f'range((0,{matrix_width}),(0,{matrix_height}))')
        return button_names[y][x]

    def _determine_bounds(self, button_names):
        assert len(button_names) > 0
        matrix_width = len(button_names[0])
        matrix_height = len(button_names)
        return matrix_width, matrix_height


class _LedColor:
    def __init__(self,
                 value=None, *,
                 lighting_mode,
                 midi_value,
                 lighting_type=Constants.LightingType.STATIC):
        self._value = value
        self._message = None
        self._lighting_mode = lighting_mode
        self._lighting_type = lighting_type
        self._midi_value = midi_value

        assert isinstance(lighting_mode, int)
        assert isinstance(lighting_type, int)

        if not value:
            self._message = self._create_reset_message(lighting_mode,
                                                       midi_value)
        elif (not isinstance(value, ColorShade)
                and not isinstance(value, str)
                and not isinstance(value, int)
                and not isinstance(value, (tuple, list))):
            raise TypeError('Must be of type ColorShade or str '
                            'or int or tuple or list.')
        elif ((isinstance(value, str)
                and not ColorShadeStore().contains(value)
                and not RgbColor.is_valid(value))
                or (isinstance(value, (tuple, list))
                    and not RgbColor.is_valid(value))):
            raise ValueError('Invalid color.')
        elif RgbColor.is_valid(value):
            colorspec = self._create_colorspec_message(value,
                                                       lighting_type,
                                                       midi_value)
            self._message = colorspec
        else:
            lighting = self._create_lighting_message(value,
                                                     lighting_mode,
                                                     midi_value)
            self._message = lighting

    @property
    def message(self):
        return self._message

    def _create_colorspec_message(self, value, lighting_type, midi_value):
        rgb_color = RgbColor(value)
        colorspec = Colorspec(ColorspecFragment(lighting_type,
                                                midi_value,
                                                rgb_color.r,
                                                rgb_color.g,
                                                rgb_color.b))
        return colorspec

    def _create_lighting_message(self, value, lighting_mode, midi_value):
        color_id = value if ColorShade.is_valid_id(value) else -1
        lighting = None
        if color_id < 0:
            color_shade = ColorShadeStore().find(value)
            color_id = (color_shade.color_id
                        if color_shade
                        else color_id)
        if color_id < 0:
            raise ValueError(f'Color ID values must be between '
                             f'{ColorShade.MIN_COLOR_ID} and '
                             f'{ColorShade.MAX_COLOR_ID}.')
        else:
            lighting = Lighting(lighting_mode,
                                midi_value,
                                color_id)
        return lighting

    def _create_reset_message(self, lighting_mode, midi_value):
        lighting = Lighting(lighting_mode,
                            midi_value,
                            0x0)
        return lighting


class ButtonFace:
    """
    A button face.

    A button face is the marking placed on the top of
    some Launchpad buttons.
    """
    UP = 'up'
    DOWN = 'down'
    LEFT = 'left'
    RIGHT = 'right'
    SESSION = 'session'
    DRUMS = 'drums'
    KEYS = 'keys'
    USER = 'user'
    LOGO = 'logo'
    SCENE_LAUNCH_1 = 'scene_launch_1'
    SCENE_LAUNCH_2 = 'scene_launch_2'
    SCENE_LAUNCH_3 = 'scene_launch_3'
    SCENE_LAUNCH_4 = 'scene_launch_4'
    SCENE_LAUNCH_5 = 'scene_launch_5'
    SCENE_LAUNCH_6 = 'scene_launch_6'
    SCENE_LAUNCH_7 = 'scene_launch_7'
    STOP_SOLO_MUTE = 'stop_solo_mute'


class ButtonGroup:
    """
    A group of buttons.
    """

    def __init__(self, launchpad,
                 layout,
                 button_names,
                 args):
        self._launchpad = launchpad
        self._layout = layout
        self._button_names = button_names
        self._buttons = self._create_buttons(launchpad,
                                             layout,
                                             button_names,
                                             args)

    def __iter__(self):
        return iter(self._buttons)

    def __repr__(self):
        names = list(map(lambda name: f"'{name}'", self.names))
        return f'ButtonGroup({names})'

    @property
    def launchpad(self):
        """
        Launchpad reference.
        """
        return self._launchpad

    @property
    def names(self):
        """
        Names of buttons in group.
        """
        return [button.name for button in self._buttons]

    def poll_for_event(self, *, interface='midi', timeout=None,
                       type=ButtonEvent.PRESS_RELEASE):
        """
        Polls for MIDI events of buttons specified
        in this group. If `timeout` is `None`, this function will
        wait indefinitely until a :class:`ButtonEvent` is
        received. If no event is received, `None` is returned.

        Keyword Args:
            interface (str): Interface to which to send message.
                (See :class:`Interface`.)
            timeout (float): Duration in seconds to wait for event to occur
            type (str): Event type.
                (Possible values: 'press', 'release', 'press|release')

        Returns:
            ButtonEvent: Button event (See :class:`ButtonEvent`).
        """
        if (not type
                or (type.lower().replace('|', '_') != ButtonEvent.PRESS_RELEASE  # noqa
                    and type.lower() != ButtonEvent.RELEASE
                    and type.lower() != ButtonEvent.PRESS)):
            raise ValueError('Not a valid event type.')

        midi_event = self._launchpad.poll_for_event(interface=interface,
                                                    timeout=timeout,
                                                    match=ButtonMatch(self._buttons,  # noqa
                                                                      type))  # noqa
        return (ButtonEvent(midi_event, self._buttons)
                if midi_event
                else None)

    def clear_event_queue(self, *, interface='midi'):
        """
        Clears event queue.

        Keyword Args:
            interface (str): Interface to clear.
                (See :class:`Interface`.)

        Raises:
            ValueError: If `interface` is invalid.
            RuntimeError: If device is closed.
        """
        self._launchpad.clear_event_queue()

    # FIXME: Too complex
    def _create_buttons(self, launchpad, layout, button_names, args):  # noqa
        buttons = []
        for arg in args:
            found = False
            for column, button_name_column in enumerate(button_names):
                if found:
                    break
                for row, button_name in enumerate(button_name_column):
                    if found:
                        break
                    elif arg and isinstance(arg, str):
                        if arg.lower() == button_name:
                            buttons.append(Button(launchpad,
                                                  layout,
                                                  button_names,
                                                  name=arg))
                            found = True
                    elif isinstance(arg, tuple):
                        if (len(arg) == 2
                                and arg[0] == row
                                and arg[1] == column):
                            buttons.append(Button(launchpad,
                                                  layout,
                                                  button_names,
                                                  x=row, y=column))
                            found = True
                    elif isinstance(arg, int):
                        buttons.append(Button(launchpad,
                                              layout,
                                              button_names,
                                              button_id=arg))
                        found = True
                    else:
                        raise ValueError(f'Invalid button "{str(arg)}".')
        return set(buttons)


class Led:
    """
    An LED on the Launchpad.
    """

    OFF = 'off'
    STATIC = 'static'
    FLASH = 'flash'
    PULSE = 'pulse'

    _LIGHTING_MODE = {
            'off': Constants.LightingMode.OFF,
            'static': Constants.LightingMode.STATIC,
            'flash': Constants.LightingMode.FLASH
    }
    _LIGHTING_TYPE = {
            'static': Constants.LightingType.STATIC,
            'flash': Constants.LightingType.FLASH,
            'pulse': Constants.LightingType.PULSE,
            'rgb': Constants.LightingType.RGB
    }

    def __init__(self, *, launchpad, button_names,
                 layout, x=-1, y=-1, name='', mode=STATIC):
        self._launchpad = launchpad
        self._button_names = button_names
        self._mode = mode
        self._x = self._y = -1

        name = (x if isinstance(x, str)
                and not name else name)
        if isinstance(x, str) and not name:
            raise ValueError('Invalid name.')
        x = -1 if isinstance(x, str) else x

        led_id = x if x >= 0 and y < 0 else -1
        x = -1 if isinstance(x, int) and led_id >= 0 else x

        coordinate = _MatrixCoordinate(launchpad=launchpad,
                                       layout=layout,
                                       button_names=button_names,
                                       name=name,
                                       coordinate_id=led_id,
                                       x=x, y=y)

        self._x = coordinate.x
        self._y = coordinate.y
        self._matrix_width = coordinate.matrix_width
        self._matrix_height = coordinate.matrix_height
        self._midi_value = coordinate.midi_value

    def __eq__(self, other):
        if not isinstance(other, Led):
            return False
        return (self.launchpad == other.launchpad
                and self.name == other.name)

    def __hash__(self):
        return hash(repr(self))

    def __repr__(self):
        return f"Led(x={self.x}, y={self.x}, name='{self.name}')"

    @property
    def id(self):
        """
        Unique ID of LED.
        """
        if not self._is_within_range():
            return -1
        return (self._y * self._matrix_height) + self._x + 1

    @property
    def x(self):
        """
        X position of LED.
        """
        return self._x

    @property
    def y(self):
        """
        Y position of LED.
        """
        return self._y

    @property
    def name(self):
        """
        Name of LED.
        """
        return (self._button_names[self.y][self.x]
                if self._is_within_range()
                else '')

    @property
    def midi_value(self):
        """
        Midi value of LED.
        """
        return self._midi_value

    @property
    def launchpad(self):
        """
        Launchpad reference.
        """
        return self._launchpad

    @property
    def color(self):
        """
        Color. Retrieving the set color is not supported.
        """
        return None

    @color.setter
    def color(self, value):
        """
        Sets the color of the LED to `value`.

        Args:
            value (ColorShade or str or int): Color value.

        Raises:
            ValueError: When invalid value is used.
            TypeError: When invalid type is used.
        """
        message = _LedColor(value,
                            lighting_mode=self._LIGHTING_MODE[self._mode],
                            lighting_type=self._LIGHTING_TYPE['rgb'],
                            midi_value=self._midi_value).message
        self.launchpad.send_message(message)

    @color.deleter
    def color(self):
        """Turns LED off."""
        self.reset()

    def reset(self):
        """Turns LED off."""
        message = _LedColor(lighting_mode=self._LIGHTING_MODE[self._mode],
                            midi_value=self._midi_value).message
        self.launchpad.send_message(message)

    def _is_within_range(self):
        return (self._x >= 0
                and self._y >= 0
                and self._x < self._matrix_width
                and self._y < self._matrix_height)


class Button:
    """
    A button on the Launchpad.
    """

    def __init__(self, launchpad,
                 layout,
                 button_names, *,
                 x=-1, y=-1,
                 name='',
                 button_id=-1):
        coordinate = _MatrixCoordinate(launchpad=launchpad,
                                       layout=layout,
                                       button_names=button_names,
                                       name=name,
                                       coordinate_id=button_id,
                                       x=x, y=y)
        self._launchpad = launchpad
        self._layout = layout
        self._button_names = button_names
        self._name = coordinate.name
        self._x = coordinate.x
        self._y = coordinate.y
        self._button_id = coordinate.id
        self._midi_value = coordinate.midi_value

    def __eq__(self, other):
        if not isinstance(other, Button):
            return False
        return (self.launchpad == other.launchpad
                and self.name == other.name)

    def __hash__(self):
        return hash(repr(self))

    def __repr__(self):
        return ("Button("
                f"name='{self.name}', "
                f"x={self.x}, "
                f"y={self.y}, "
                f"id={self.id})")

    @property
    def launchpad(self):
        """
        Launchpad reference.
        """
        return self._launchpad

    @property
    def x(self):
        """
        X position of button.
        """
        return self._x

    @property
    def y(self):
        """
        Y position of button.
        """
        return self._y

    @property
    def name(self):
        """
        Name of button.
        """
        return self._name

    @property
    def midi_value(self):
        """
        MIDI value of button.
        """
        return self._midi_value

    @property
    def id(self):
        """
        Unique ID of button.
        """
        return self._button_id

    @property
    def parent(self):
        """
        Parent of button, either 'grid' or 'panel'.
        """
        match = re.match('^\\d+x\\d+$', self._name)
        if match:
            return 'grid'
        return 'panel' if self._button_id else ''

    @property
    def led(self):
        """
        LED of button.
        """
        return Led(launchpad=self._launchpad,
                   layout=self._layout,
                   button_names=self._button_names,
                   x=self._x,
                   y=self._y)

    @property
    def layout(self):
        """
        Button layout.
        """
        return self._layout


class Panel(Matrix):
    """
    Panel of Launchpad.

    The panel represents the 9x9 matrix of pad buttons
    on the surface of the Launchpad.
    """
    PROG = 'prog'
    CUSTOM = 'custom'
    _PROG_MODE_MIDI_LAYOUT = [
            [0x5b, 0x5c, 0x5d, 0x5e, 0x5f, 0x60, 0x61, 0x62, 0x63],
            [0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59],
            [0x47, 0x48, 0x49, 0x4a, 0x4b, 0x4c, 0x4d, 0x4e, 0x4f],
            [0x3d, 0x3e, 0x3f, 0x40, 0x41, 0x42, 0x43, 0x44, 0x45],
            [0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3a, 0x3b],
            [0x29, 0x2a, 0x2b, 0x2c, 0x2d, 0x2e, 0x2f, 0x30, 0x31],
            [0x1f, 0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27],
            [0x15, 0x16, 0x17, 0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d],
            [0x0b, 0x0c, 0x0d, 0x0e, 0x0f, 0x10, 0x11, 0x12, 0x13]
    ]
    _CUSTOM_MODE_MIDI_LAYOUT = [
            [0x5b, 0x5c, 0x5d, 0x5e, 0x5f, 0x60, 0x61, 0x62, 0x63],
            [0x40, 0x41, 0x42, 0x43, 0x60, 0x61, 0x62, 0x63, 0x59],
            [0x3c, 0x3d, 0x3e, 0x3f, 0x5c, 0x5d, 0x5e, 0x5f, 0x4f],
            [0x38, 0x39, 0x3a, 0x3b, 0x58, 0x59, 0x5a, 0x5b, 0x45],
            [0x34, 0x35, 0x36, 0x37, 0x54, 0x55, 0x56, 0x57, 0x3b],
            [0x30, 0x31, 0x32, 0x33, 0x50, 0x51, 0x52, 0x53, 0x31],
            [0x2c, 0x2d, 0x2e, 0x2f, 0x4c, 0x4d, 0x4e, 0x4f, 0x27],
            [0x28, 0x29, 0x2a, 0x2b, 0x48, 0x49, 0x4a, 0x4b, 0x1d],
            [0x24, 0x25, 0x26, 0x27, 0x44, 0x45, 0x46, 0x47, 0x13]
    ]

    _BUTTON_NAMES = [
            ['up', 'down', 'left', 'right', 'session', 'drums', 'keys', 'user', 'logo'],  # noqa
            ['0x0', '1x0', '2x0', '3x0', '4x0', '5x0', '6x0', '7x0', 'scene_launch_1'],  # noqa
            ['0x1', '1x1', '2x1', '3x1', '4x1', '5x1', '6x1', '7x1', 'scene_launch_2'],  # noqa
            ['0x2', '1x2', '2x2', '3x2', '4x2', '5x2', '6x2', '7x2', 'scene_launch_3'],  # noqa
            ['0x3', '1x3', '2x3', '3x3', '4x3', '5x3', '6x3', '7x3', 'scene_launch_4'],  # noqa
            ['0x4', '1x4', '2x4', '3x4', '4x4', '5x4', '6x4', '7x4', 'scene_launch_5'],  # noqa
            ['0x5', '1x5', '2x5', '3x5', '4x5', '5x5', '6x5', '7x5', 'scene_launch_6'],  # noqa
            ['0x6', '1x6', '2x6', '3x6', '4x6', '5x6', '6x6', '7x6', 'scene_launch_7'],  # noqa
            ['0x7', '1x7', '2x7', '3x7', '4x7', '5x7', '6x7', '7x7', 'stop_solo_mute']   # noqa
    ]

    def __init__(self, launchpad):
        self._launchpad = launchpad

    def __eq__(self, other):
        if not isinstance(other, Panel):
            return False
        return self.launchpad == other.launchpad

    def __repr__(self):
        return f'Panel({self.width}x{self.height})'

    @Matrix.launchpad.getter
    def launchpad(self):
        """
        Launchpad reference.
        """
        return self._launchpad

    @Matrix.width.getter
    def width(self):
        """
        Max X.
        """
        return len(Panel._BUTTON_NAMES[0])

    @Matrix.height.getter
    def height(self):
        """
        Max Y.
        """
        return len(Panel._BUTTON_NAMES)

    @property
    def max_id(self):
        """
        Max ID.
        """
        return self.width * self.height

    @Matrix.max_x.getter
    def max_x(self):
        """
        Max X.
        """
        return (self.width - 1
                if self.width >= 0
                else -1)

    @Matrix.max_y.getter
    def max_y(self):
        """
        Max Y.
        """
        return (self.height - 1
                if self.height >= 0
                else -1)

    def led(self, x=-1, y=-1, *, name='', layout=PROG, mode=Led.STATIC):
        """
        Returns an LED.

        Args:
            x (int): X position of LED.
            y (int): Y position of LED.

        Keyword Args:
            name (str): Name of LED.
            layout (Layout): Layout of buttons.
            mode (str): Lighting mode.
        """
        return Led(launchpad=self._launchpad,
                   button_names=Panel._BUTTON_NAMES,
                   layout=(Panel._CUSTOM_MODE_MIDI_LAYOUT
                           if layout == Panel.CUSTOM
                           else Panel._PROG_MODE_MIDI_LAYOUT),
                   x=x, y=y,
                   name=name, mode=mode)

    # FIXME: Too complex
    def led_range(self, *,  # noqa
                  layout=PROG,
                  mode=Led.STATIC,
                  rotation=0,
                  flip_axis='',
                  region=None):
        """
        Returns an immutable sequence of LEDs.

        Keyword Args:
            layout (Layout): Layout of buttons.
            mode (str): Lighting mode.
            rotation (int): Rotation angle in degrees.
                (Possible values: 0, 90, 180, 270, -90)

        Yields:
            Led: Sequence of LEDs
        """
        if region and not isinstance(region, Region):
            raise TypeError("'region' must be of type 'Region'.")
        elif rotation and not isinstance(rotation, int):
            raise TypeError("'rotation' must be of type 'int'.")
        elif flip_axis and not isinstance(flip_axis, str):
            raise TypeError("'flip_axis' must be of type 'str'.")

        if region:
            for name in region.button_names:
                yield self.led(name=name, layout=layout, mode=mode)
        elif rotation:
            for led in _MatrixTransform(self, layout, mode).rotated_led_range(rotation,  # noqa
                                                                              flip_axis=flip_axis):  # noqa
                yield led
        elif not rotation and flip_axis:
            for led in _MatrixTransform(self, layout, mode).flipped_led_range(flip_axis):  # noqa
                yield led
        else:
            for led_id in range(self.max_id):
                yield self.led(led_id, layout=layout, mode=mode)

    def buttons(self, *args, layout=PROG):
        """
        Returns a :class:`ButtonGroup`.

        Args:
            args (list): Button names, button IDs
                or button XY-pairs

        Keyword Args:
            layout (Layout): Layout of buttons.
        """
        args = (args
                if len(args) > 0
                else [button_name
                      for button_row in Panel._BUTTON_NAMES
                      for button_name in button_row])
        return ButtonGroup(launchpad=self._launchpad,
                           layout=(Panel._CUSTOM_MODE_MIDI_LAYOUT
                                   if layout == Panel.CUSTOM
                                   else Panel._PROG_MODE_MIDI_LAYOUT),
                           button_names=Panel._BUTTON_NAMES,
                           args=list(args))


class Grid(Matrix):
    """
    Grid of Launchpad.

    The grid represents the 8x8 grid of white, faceless
    buttons of the Launchpad.
    """
    PROG = 'prog'
    CUSTOM = 'custom'
    _PROG_MODE_MIDI_LAYOUT = [
            [0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57, 0x58],
            [0x47, 0x48, 0x49, 0x4a, 0x4b, 0x4c, 0x4d, 0x4e],
            [0x3d, 0x3e, 0x3f, 0x40, 0x41, 0x42, 0x43, 0x44],
            [0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3a],
            [0x29, 0x2a, 0x2b, 0x2c, 0x2d, 0x2e, 0x2f, 0x30],
            [0x1f, 0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26],
            [0x15, 0x16, 0x17, 0x18, 0x19, 0x1a, 0x1b, 0x1c],
            [0x0b, 0x0c, 0x0d, 0x0e, 0x0f, 0x10, 0x11, 0x12]
    ]
    _CUSTOM_MODE_MIDI_LAYOUT = [
            [0x40, 0x41, 0x42, 0x43, 0x60, 0x61, 0x62, 0x63],
            [0x3c, 0x3d, 0x3e, 0x3f, 0x5c, 0x5d, 0x5e, 0x5f],
            [0x38, 0x39, 0x3a, 0x3b, 0x58, 0x59, 0x5a, 0x5b],
            [0x34, 0x35, 0x36, 0x37, 0x54, 0x55, 0x56, 0x57],
            [0x30, 0x31, 0x32, 0x33, 0x50, 0x51, 0x52, 0x53],
            [0x2c, 0x2d, 0x2e, 0x2f, 0x4c, 0x4d, 0x4e, 0x4f],
            [0x28, 0x29, 0x2a, 0x2b, 0x48, 0x49, 0x4a, 0x4b],
            [0x24, 0x25, 0x26, 0x27, 0x44, 0x45, 0x46, 0x47]
    ]
    _BUTTON_NAMES = [
            ['0x0', '1x0', '2x0', '3x0', '4x0', '5x0', '6x0', '7x0'],
            ['0x1', '1x1', '2x1', '3x1', '4x1', '5x1', '6x1', '7x1'],
            ['0x2', '1x2', '2x2', '3x2', '4x2', '5x2', '6x2', '7x2'],
            ['0x3', '1x3', '2x3', '3x3', '4x3', '5x3', '6x3', '7x3'],
            ['0x4', '1x4', '2x4', '3x4', '4x4', '5x4', '6x4', '7x4'],
            ['0x5', '1x5', '2x5', '3x5', '4x5', '5x5', '6x5', '7x5'],
            ['0x6', '1x6', '2x6', '3x6', '4x6', '5x6', '6x6', '7x6'],
            ['0x7', '1x7', '2x7', '3x7', '4x7', '5x7', '6x7', '7x7']
    ]

    def __init__(self, launchpad):
        self._launchpad = launchpad

    def __eq__(self, other):
        if not isinstance(other, Grid):
            return False
        return self.launchpad == other.launchpad

    def __repr__(self):
        return f'Grid({self.width}x{self.height})'

    @Matrix.launchpad.getter
    def launchpad(self):
        """
        Launchpad reference.
        """
        return self._launchpad

    @Matrix.width.getter
    def width(self):
        """
        Max X.
        """
        return len(Grid._BUTTON_NAMES[0])

    @Matrix.height.getter
    def height(self):
        """
        Max Y.
        """
        return len(Grid._BUTTON_NAMES)

    @property
    def max_id(self):
        """
        Max ID.
        """
        return self.width * self.height

    @Matrix.max_x.getter
    def max_x(self):
        """
        Max X.
        """
        return (self.width - 1
                if self.width >= 0
                else -1)

    @Matrix.max_y.getter
    def max_y(self):
        """
        Max Y.
        """
        return (self.height - 1
                if self.height >= 0
                else -1)

    def led(self, x=-1, y=-1, *, name='', layout=PROG, mode=Led.STATIC):
        """
        Returns an LED.

        Args:
            x (int): X position of LED.
            y (int): Y position of LED.

        Keyword Args:
            name (str): Name of LED.
            layout (Layout): Layout of buttons.
            mode (str): Lighting mode.
        """
        return Led(launchpad=self._launchpad,
                   button_names=Grid._BUTTON_NAMES,
                   layout=(Grid._CUSTOM_MODE_MIDI_LAYOUT
                           if layout == Grid.CUSTOM
                           else Grid._PROG_MODE_MIDI_LAYOUT),
                   x=x, y=y,
                   name=name, mode=mode)

    # FIXME: Too complex
    def led_range(self, *,  # noqa
                  layout=PROG,
                  mode=Led.STATIC,
                  rotation=0,
                  flip_axis='',
                  region=None):
        """
        Returns an immutable sequence of LEDs.

        Keyword Args:
            layout (Layout): Layout of buttons.
            mode (str): Lighting mode.
            rotation (int): Rotation angle in degrees.
                (Possible values: 0, 90, 180, 270, -90)

        Yields:
            Led: Sequence of LEDs
        """
        if region and not isinstance(region, Region):
            raise TypeError("'region' must be of type 'Region'.")
        elif rotation and not isinstance(rotation, int):
            raise TypeError("'rotation' must be of type 'int'.")
        elif flip_axis and not isinstance(flip_axis, str):
            raise TypeError("'flip_axis' must be of type 'str'.")

        if region:
            for name in region.button_names:
                yield self.led(name=name, layout=layout, mode=mode)
        elif rotation:
            for led in _MatrixTransform(self, layout, mode).rotated_led_range(rotation,  # noqa
                                                                              flip_axis=flip_axis):  # noqa
                yield led
        elif not rotation and flip_axis:
            for led in _MatrixTransform(self, layout, mode).flipped_led_range(flip_axis):  # noqa
                yield led
        else:
            for led_id in range(self.max_id):
                yield self.led(led_id, layout=layout, mode=mode)

    def buttons(self, *args, layout=PROG):
        """
        Returns a :class:`ButtonGroup`.

        Args:
            args (list): Button names, button IDs
                or button XY-pairs

        Keyword Args:
            layout (Layout): Layout of buttons.
        """
        args = (args
                if len(args) > 0
                else [button_name
                      for button_row in Grid._BUTTON_NAMES
                      for button_name in button_row])
        return ButtonGroup(launchpad=self._launchpad,
                           layout=(Grid._CUSTOM_MODE_MIDI_LAYOUT
                                   if layout == Grid.CUSTOM
                                   else Grid._PROG_MODE_MIDI_LAYOUT),
                           button_names=Grid._BUTTON_NAMES,
                           args=list(args))



import math
from ._parser import BitmapConfig


class RawBitmapMatrix:
    def __init__(self, raw_bitmap):
        self._raw_bitmap = raw_bitmap
        self._raw_matrix = self._load(raw_bitmap)

    def __repr__(self):
        return ('RawBitmapMatrix('
                f'bit={self._raw_matrix})')

    def bit(self, x, y):
        return self._raw_matrix[y][x]

    def rotated_range(self, angle, *, flip_axis=''):
        assert angle, 'Angle cannot be zero.'
        angle = self._normalize_angle(angle)
        angle = self._flip_angle(angle)
        angle_rad = math.radians(angle)
        for index, bit in enumerate(self._raw_bitmap):
            x = int(index / self._raw_bitmap.word_count)
            y = int(index % self._raw_bitmap.word_count)
            max_x = max_y = self._raw_bitmap.word_count - 1
            rotated_x = round((y * math.sin(angle_rad)) + (x * math.cos(angle_rad)))  # noqa
            rotated_y = round((y * math.cos(angle_rad)) - (x * math.sin(angle_rad)))  # noqa
            x_flip_required = y_flip_required = False
            if angle == 90:
                rotated_y = min(max_y + rotated_y, max_y)
                x_flip_required = (FlipAxis.X in flip_axis)
                y_flip_required = (FlipAxis.Y in flip_axis)
            elif angle == 180:
                rotated_y = min(max_y + rotated_y, max_y)
                rotated_x = min(max_x + rotated_x, max_x)
                x_flip_required = FlipAxis.Y in flip_axis
                y_flip_required = FlipAxis.X in flip_axis
            elif angle == 270:
                rotated_x = min(max_x + rotated_x, max_x)
                x_flip_required = (FlipAxis.X in flip_axis)
                y_flip_required = (FlipAxis.Y in flip_axis)
            if flip_axis:
                flipped_x = (self._flip_axis(rotated_x)
                             if x_flip_required
                             else rotated_x)
                flipped_y = (self._flip_axis(rotated_y)
                             if y_flip_required
                             else rotated_y)
                yield self.bit(flipped_x, flipped_y)
            else:
                yield self.bit(rotated_x, rotated_y)

    def flipped_range(self, axis):
        for index, bit in enumerate(self._raw_bitmap):
            x = int(index / self._raw_bitmap.word_count)
            y = int(index % self._raw_bitmap.word_count)
            flipped_x = (self._flip_axis(x)
                         if FlipAxis.Y in axis
                         else x)
            flipped_y = (self._flip_axis(y)
                         if FlipAxis.X in axis
                         else y)
            yield self.bit(flipped_x, flipped_y)

    def _blank_matrix(self):
        matrix = []
        for _ in range(self._raw_bitmap.word_count):
            matrix.append([])
            for _ in range(self._raw_bitmap.word_count):
                matrix[-1].append(0)
        return matrix

    def _load(self, raw_bitmap):
        matrix = self._blank_matrix()
        for index, bit in enumerate(raw_bitmap):
            x = int(index % raw_bitmap.word_count)
            y = int(index / raw_bitmap.word_count)
            matrix[x][y] = bit

        return matrix

    def _normalize_angle(self, angle):
        if angle and (angle % 90) == 0 and abs(angle) > 270:
            return angle % 360
        assert (round(abs(angle)) in (0, 90, 180, 270)), \
               'Angle must be a multiple of 90.'
        return angle

    def _flip_axis(self, value):
        max_value = self._raw_bitmap.word_count - 1
        return max_value - value

    def _flip_angle(self, angle):
        if angle == -90 or angle == 270:
            return 90
        elif angle == -180:
            return 180
        elif angle == 90 or angle == -270:
            return 270
        return angle


class RawBitmap:
    def __init__(self, bitmap_data, config_data=None):
        assert isinstance(bitmap_data, list)
        self._data = bitmap_data
        self._config = BitmapConfig(config_data)

    def __iter__(self):
        for word in self._data:
            bitmask = 1
            for _ in range(self.word_count):
                bit = 1 if (word & bitmask) else 0
                yield bit
                bitmask = bitmask << 1

    def __repr__(self):
        return f"RawBitmap('{self._data}')"

    def __str__(self):
        bit_string = ''
        for index, bit in enumerate(self, start=1):
            bit_string += str(bit)
            bit_string += ('\n'
                           if index % self.word_count == 0
                           else ' ')
        return bit_string

    @property
    def data(self):
        return self._data

    @property
    def word_count(self):
        return len(self._data)

    @property
    def config(self):
        return self._config

    def render(self):
        pass


class Offset:
    def __init__(self, x=0, y=0):
        self._x = x
        self._y = y

    def __repr__(self):
        return f'Offset({self.x}, {self.y})'

    def __str__(self):
        return f'({self.x}, {self.y})'

    @property
    def x(self):
        return self._x

    @property
    def y(self):
        return self._y


class Framerate:
    def __init__(self,
                 movie_ref,
                 movie):
        self._movie_ref = movie_ref
        self._movie = movie

    def set(self, framerate):
        self._movie.framerate = framerate
        return self._movie_ref

    def __int__(self):
        return self._movie.framerate

    def __repr__(self):
        return f"Framerate({self._movie.framerate})"

    def __str__(self):
        return repr(self)


class FlipAxis:
    X = 'x'
    Y = 'y'
    XY = 'xy'


class ScrollDirection:
    LEFT = 'left'
    RIGHT = 'right'



"""
Display characters typed with a QWERTY keyboard on the Launchpad's surface.
"""

from lpminimk3 import Mode, find_launchpads
from lpminimk3.graphics import Text
import sys
import termios
import tty


def getchar():
    fd = sys.stdin.fileno()
    old_settings = termios.tcgetattr(fd)
    try:
        tty.setraw(sys.stdin.fileno())
        ch = sys.stdin.read(1)
    finally:
        termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
    return ch


def render_character(lp):
    while True:
        c = getchar()
        if ord(c) == 0x03:
            sys.exit()
        else:
            lp.grid.render(Text(c))


def main():
    lp = find_launchpads()[0]  # Get the first available launchpad
    lp.open()  # Open device for reading and writing on MIDI interface (by default)  # noqa

    lp.mode = Mode.PROG  # Switch to the programmer mode

    print('Press a key on your keyboard to display a character on your Launchpad.\n'  # noqa
          'Press Ctrl+C to quit.\n')
    render_character(lp)  # Render character on Launchpad's surface


if __name__ == '__main__':
    main()



import os
from ..colors import ColorPalette
from ._renderable import Renderable,\
                         RenderableColor,\
                         String,\
                         TextScroll,\
                         MovieRoll,\
                         Bitmap as _Bitmap,\
                         Movie as _Movie
from ._utils import ScrollDirection,\
                    FlipAxis,\
                    Framerate
from ._renderable import Frame  # noqa


class Text(Renderable):
    """
    Text renderable on the Launchpad's surface. The text
    rendered can be transformed by calling any of the
    transformation methods of this class.


    EXAMPLES
    Render text on Launchpad's surface:
        lp.grid.render(Text('A'))

    Print the text in the console:
        Text('A').print()

    Scroll text and render text on Launchpad's surface:
        lp.grid.render(Text(' Hello, world!').scroll())

    Scroll text and render text in the console:
        Text(' Hello, world!').print()

    Set the foreground color to "blue", rotate the text
    anticlockwise by 90 degrees, flip the text on its 'X'
    axis and scroll once:
        lp.grid.render(Text(' Hello, world!')
                       .fg_color.set('blue')
                       .rotate(-90)
                       .flip()
                       .scroll(count=1))
    """
    def __init__(self,
                 text='', *,
                 fg_color=ColorPalette.Red.SHADE_4,
                 bg_color=None):
        if isinstance(text, Text):
            text = text.text
            fg_color = text.fg_color
            bg_color = text.bg_color
        elif not isinstance(text, str):
            raise TypeError('Must be of type str.')

        self._text = text
        self._string = String(text,
                              fg_color=RenderableColor(self,
                                                       fg_color),
                              bg_color=RenderableColor(self,
                                                       bg_color))

    def __repr__(self):
        return f"Text('{self._text}')"

    def __str__(self):
        return self._text

    def __len__(self):
        return len(self._text)

    @Renderable.bits.getter
    def bits(self):
        """
        Text as bits (i.e. 0s and 1s).
        """
        return self._string.bits

    @Renderable.word_count.getter
    def word_count(self):
        """
        Number of (bit)words.
        """
        return self._string.word_count

    @property
    def fg_color(self):
        """
        Foreground color.
        """
        return self._string.fg_color

    @fg_color.setter
    def fg_color(self, color):
        self._string.fg_color = RenderableColor(self, color)

    @property
    def bg_color(self):
        """
        Background color.
        """
        return self._string.bg_color

    @bg_color.setter
    def bg_color(self, color):
        self._string.bg_color = RenderableColor(self, color)

    def print(self, one='X', zero=' '):
        """
        Prints text to the screen as bits, used to give a preview of
        how text would be rendered on the Launchpad's surface.

        Keyword Args:
            one (str): Character to represent bit 1.
            zero (str): Character to represent bit 0.

        Returns:
            Text: `Text` reference.
        """
        self._string.print(one=one, zero=zero)

    def shift(self, count=1, *, circular=True):
        """
        Shifts text to the left if `count` is positive, or to the
        right if `count` is negative.

        Args:
            count (int): Number of times to shift.

        Keyword Args:
            circular (bool): If True, enable text wrapping; else
                diable text wrapping.

        Returns:
            Text: `Text` reference.

        Raises:
            ValueError: When invalid value is set.
        """
        if not isinstance(count, int):
            raise ValueError("'count' must be 'int'.")
        if count > 0:
            self._string.shift_right(count=count, circular=circular)
        elif count < 0:
            self._string.shift_left(count=int(abs(count)), circular=circular)
        return self

    def rotate(self, angle):
        """
        Rotates rendered text by `angle` degrees. All values
        must be a multiples of 90.

        Args:
            angle (int): Angle of rotation in degrees.
                A negative `angle` rotates Text in an anticlockwise
                direction; a positive `angle` clockwise.

        Returns:
            Text: `Text` reference.
        """
        self._string.rotate(angle)
        return self

    def scroll(self, *,
               period=.05,
               direction=ScrollDirection.LEFT,
               cycle_func=None,
               count=None,
               timeout=None):
        """
        Scrolls rendered text, shifting every `period` seconds in the
        `direction` direction; scrolls indefinitely if `count` is not
        set.

        Keyword Args:
            period (float): Delay before every text render call.
            direction (str): Direction of scroll, either left or right.
                (See :class:`ScrollDirection`.)
            cycle_func (callable): Function to be called right after
                text render call but before period delay. The function
                signature is:
                    cycle_func(fraction, launchpad)
                where `fraction` is the fraction of the scroll (ranging
                from 0 to 1) and `launchpad` is the Launchpad reference.
            count (int): Number of complete scrolls. Scrolls
                indefinitely if `count` and `timeout` are not set.
            timeout (float): Duration in seconds for scroll. Scrolls
                indefinitely if `count` and `timeout` are not set.
                If both `timeout` and `count` are set, the scroll will
                end depending on which value is reached first.

        Returns:
            Text: `Text` reference.

        Raises:
            ValueError: When invalid value is set.
        """
        if timeout and timeout <= period:
            raise ValueError("'timeout' must be greater than "
                             "or equal to the period.")
        if cycle_func and not callable(cycle_func):
            raise ValueError("'cycle_func' must be callable.")
        self._string.text_scroll = TextScroll(self._text,
                                              period,
                                              direction,
                                              cycle_func=cycle_func,
                                              timeout=timeout,
                                              count=count)
        return self

    def flip(self, axis=FlipAxis.X):
        """
        Flips text on `axis` axis.

        Args:
            axis (str): Axis to flip on (X, Y or both).
                (See :class:`FlipAxis`.)

        Returns:
            Text: `Text` reference.
        """
        self._string.flip_axis = axis
        return self

    def swap_colors(self):
        """
        Swaps foreground and background colors.

        Returns:
            Text: `Text` reference.
        """
        self._string.fg_color, self._string.bg_color = self._string.bg_color, self._string.fg_color  # noqa
        return self

    def render(self, matrix):
        """
        Renders text on matrix `matrix`. Used for internal purposes only.
        """
        self._string.render(matrix)


class Bitmap(Renderable):
    def __init__(self,
                 filename, *,
                 fg_color=None,
                 bg_color=None):
        if not isinstance(filename, str):
            raise TypeError('Must be of type str.')
        if not filename:
            raise ValueError('No filename set.')
        self._bitmap = _Bitmap(filename,
                               fg_color=RenderableColor(self,
                                                        fg_color),
                               bg_color=RenderableColor(self,
                                                        bg_color))

    def __repr__(self):
        return ("Bitmap("
                f"filename='{os.path.basename(self.filename)}')")

    def __str__(self):
        return repr(self)

    @Renderable.bits.getter
    def bits(self):
        return self._bitmap.bits

    @Renderable.word_count.getter
    def word_count(self):
        return self._bitmap.word_count

    @property
    def filename(self):
        return self._bitmap.filename

    @property
    def fg_color(self):
        return self._bitmap.fg_color

    @fg_color.setter
    def fg_color(self, color):
        self._bitmap.fg_color = RenderableColor(self, color)

    @property
    def bg_color(self):
        return self._bitmap.bg_color

    @bg_color.setter
    def bg_color(self, color):
        self._bitmap.bg_color = RenderableColor(self, color)

    def render(self, matrix):
        self._bitmap.render(matrix)

    def print(self, *,
              one='X',
              zero=' '):
        self._bitmap.print(one=one, zero=zero)

    def swap_colors(self):
        self._bitmap.fg_color, self._bitmap.bg_color = self._bitmap.bg_color, self._bitmap.fg_color  # noqa
        return self


class Movie(Renderable):
    def __init__(self,
                 filename, *,
                 fg_color=None,
                 bg_color=None):
        if not isinstance(filename, str):
            raise TypeError('Must be of type str.')
        if not filename:
            raise ValueError('No filename set.')
        self._movie = _Movie(filename,
                             fg_color=RenderableColor(self,
                                                      fg_color),
                             bg_color=RenderableColor(self,
                                                      bg_color))

    def __repr__(self):
        return ("Movie("
                f"filename='{os.path.basename(self.filename)}')")

    def __str__(self):
        return repr(self)

    @Renderable.bits.getter
    def bits(self):
        return self._movie.bits

    @Renderable.word_count.getter
    def word_count(self):
        return self._movie.word_count

    @property
    def filename(self):
        return self._movie.filename

    @property
    def fg_color(self):
        return self._movie.fg_color

    @fg_color.setter
    def fg_color(self, color):
        self._movie.fg_color = RenderableColor(self, color)

    @property
    def bg_color(self):
        return self._movie.bg_color

    @bg_color.setter
    def bg_color(self, color):
        self._movie.bg_color = RenderableColor(self, color)

    @property
    def framerate(self):
        return Framerate(self, self._movie)

    @property
    def frames(self):
        return self._movie.frames

    @property
    def first_frame(self):
        return self._movie.first_frame

    @property
    def last_frame(self):
        return self._movie.last_frame

    @property
    def position(self):
        return self._movie.position

    @position.setter
    def position(self, position):
        self._movie.position = position

    def render(self, matrix):
        self._movie.render(matrix)

    def play(self, *,
             count=None,
             cycle_func=None):
        if cycle_func and not callable(cycle_func):
            raise ValueError("'cycle_func' must be callable.")
        self._movie.roll = MovieRoll(self._movie,
                                     int(self.framerate),
                                     cycle_func=cycle_func,
                                     count=count)
        return self

    def skip(self, frame_count=1):
        return self._movie.skip(frame_count, ref=self)

    def print(self, *,
              one='X',
              zero=' '):
        self._movie.print(one=one, zero=zero)

    def swap_colors(self):
        self._movie.fg_color, self._movie.bg_color = self._movie.bg_color, self._movie.fg_color  # noqa
        return self



"""
Print a typed string on the console.
"""

from lpminimk3.graphics import Text


def main():
    try:
        text_to_print = ''
        while not text_to_print:
            text_to_print = input('Enter text to print '
                                  '(Press Ctrl+C to quit): ')
        text_to_print += ' '
        Text(text_to_print).scroll().print()
    except KeyboardInterrupt:
        print('\n')


if __name__ == '__main__':
    main()



from rtmidi import MidiOut, MidiIn
from ._core.components import Grid, Panel
from ._core.utils import SystemMidiPortParser, \
                          MidiPort, MidiClient, \
                          Interface, Mode, Layout # noqa
from .midi_messages import SysExMessages


class LaunchpadMiniMk3:
    """
    A Launchpad Mini MK3 device.
    """

    def __init__(self, midi_client):
        self._midi_client = midi_client

    def __eq__(self, other):
        if not isinstance(other, LaunchpadMiniMk3):
            return False
        return self.id == other.id

    def __repr__(self):
        return f'LaunchpadMiniMk3(id={self.id})'

    def is_open(self):
        """
        Returns `True` if the device is open, otherwise
        returns `False`. An open device is a device that can
        be read from and/or written to.
        """
        return self._midi_client.is_open()

    def open(self, interface=Interface.MIDI):
        """
        Opens interface `interface` on this device for reading
        and writing.

        Args:
            interface (str): Interface to open.
                (See :class:`Interface`.)

        Raise:
            ValueError: If `interface` is invalid.
        """
        self._midi_client.open(interface)
        self.interface = interface

    def close(self):
        """
        Closes this device.
        """
        self._midi_client.close()

    def send_message(self, msg, *, interface=Interface.MIDI):
        """
        Sends a MIDI message `msg` to the `interface` of
        this device.

        Args:
            message (list): MIDI message

        Keyword Args:
            interface (str): Interface to which to send message.
                (See :class:`Interface`.)

        Raises:
            ValueError: If `interface` is invalid.
            RuntimeError: If device is closed.
        """
        if self.is_open():
            if interface == Interface.DAW:
                self.daw_out_port.send_message(msg)
            elif interface == Interface.MIDI:
                self.midi_out_port.send_message(msg)
            else:
                raise ValueError('Must be a valid Interface')
        else:
            raise RuntimeError('Port closed.')

    def poll_for_event(self, *, interface=Interface.MIDI,
                       timeout=5, match=None):
        """
        Polls for a MIDI event from interface `interface` either until
        `timeout` is reached or `match` is found.

        Keyword Args:
            interface (str): Interface to poll.
                (See :class:`Interface`.)
            timeout (float): Duration in seconds to wait for event
                before returning `None`.
            match (Match): Match criterion. (See :class:`Match`).

        Returns:
            MidiEvent: MIDI event received. (See :class:`MidiEvent`).

        Raises:
            ValueError: If `interface` is invalid.
            RuntimeError: If this device is closed.
        """
        if self.is_open():
            if interface == Interface.DAW:
                return self.daw_in_port.poll_for_event(timeout=timeout,
                                                       match=match)
            elif interface == Interface.MIDI:
                return self.midi_in_port.poll_for_event(timeout=timeout,
                                                        match=match)
            else:
                raise ValueError('Must be a valid Interface.')
        else:
            raise RuntimeError('Port closed.')

    def clear_event_queue(self, *, interface=Interface.MIDI):
        """
        Clears MIDI event queue for interface `interface`.

        Keyword Args:
            interface (str): Interface to clear.
                (See :class:`Interface`.)

        Raises:
            ValueError: If `interface` is invalid.
            RuntimeError: If device is closed.
        """
        if self.is_open():
            if interface == Interface.DAW:
                return self.daw_in_port.clear_event_queue()
            elif interface == Interface.MIDI:
                return self.midi_in_port.clear_event_queue()
            else:
                raise ValueError('Must be a valid Interface.')
        else:
            raise RuntimeError('Port closed.')

    @property
    def id(self):
        """
        Unique ID of this device.
        """
        return self._midi_client.client_number

    @property
    def daw_in_port(self):
        """
        DAW interface MIDI-in port.
        Used for internal purposes.
        """
        return self._midi_client.daw_in_port

    @property
    def daw_out_port(self):
        """
        DAW interface MIDI-out port.
        Used for internal purposes.
        """
        return self._midi_client.daw_out_port

    @property
    def midi_in_port(self):
        """
        MIDI interface MIDI-in port.
        Used for internal purposes.
        """
        return self._midi_client.midi_in_port

    @property
    def midi_out_port(self):
        """
        MIDI interface MIDI-out port.
        Used for internal purposes.
        """
        return self._midi_client.midi_out_port

    @property
    def interface(self):
        """
        Interface of Launchpad. (See :class:`Interface`.)
        """
        self.send_message(SysExMessages.Interfaces.READBACK)
        return Interface(self.poll_for_event())

    @interface.setter
    def interface(self, value):
        if value.lower() == Interface.MIDI:
            self.send_message(SysExMessages.Interfaces.MIDI)
        elif value.lower() == Interface.DAW:
            self.send_message(SysExMessages.Interfaces.DAW)
        else:
            raise ValueError('Invalid interface set.')

    @property
    def panel(self):
        """
        Panel of Launchpad; represents the 9x9 array of buttons,
        with face buttons included. (See :class:`Panel`.)
        """
        return Panel(self)

    @property
    def grid(self):
        """
        Grid of Launchpad; represents the 8x8 array of white,
        faceless buttons. (See :class:`Grid`.)
        """
        return Grid(self)

    @property
    def mode(self):
        """
        Mode of Launchpad. (See :class:`Mode`.)
        """
        self.send_message(SysExMessages.Modes.READBACK)
        return Mode(self.poll_for_event())

    @mode.setter
    def mode(self, value):
        if value.lower() == Mode.LIVE:
            self.send_message(SysExMessages.Modes.LIVE)
        elif value.lower() == Mode.PROG:
            self.send_message(SysExMessages.Modes.PROG)
        else:
            raise ValueError('Invalid mode set.')

    @property
    def layout(self):
        """
        Layout of Launchpad. (See :class:`Layout`.)
        """
        self.send_message(SysExMessages.Layouts.READBACK)
        return Layout(self.poll_for_event())

    @layout.setter
    def layout(self, value):
        if value.lower() == Layout.SESSION:
            self.send_message(SysExMessages.Layouts.SESSION)
        elif value.lower() == Layout.CUSTOM_1:
            self.send_message(SysExMessages.Layouts.CUSTOM_1)
        elif value.lower() == Layout.CUSTOM_2:
            self.send_message(SysExMessages.Layouts.CUSTOM_2)
        elif value.lower() == Layout.CUSTOM_3:
            self.send_message(SysExMessages.Layouts.CUSTOM_3)
        elif value.lower() == Layout.DAW_FADERS:
            self.send_message(SysExMessages.Layouts.DAW_FADERS)
        elif value.lower() == Layout.PROG:
            self.send_message(SysExMessages.Layouts.PROG)
        else:
            raise ValueError('Invalid layout set.')

    def device_inquiry(self):
        """
        Sends a "device inquiry" message to this device.

        Returns:
            MidiEvent: MIDI event received.
        """
        self.mode = Mode.PROG
        self.send_message(SysExMessages.DEVICE_INQUIRY)
        return self.poll_for_event()


def find_launchpads():
    """
    Searches for connected Launchpad Mini MK3 devices.

    Returns:
        list of LaunchpadMiniMk3: List of found devices.
    """
    midi_out = MidiOut()
    out_ports = midi_out.get_ports()
    midi_in = MidiIn()
    in_ports = midi_in.get_ports()
    parser = SystemMidiPortParser(in_ports, out_ports)

    found_launchpads = []
    for client_data in parser.found_clients:
        midi_client = MidiClient(client_data.client_name,
                                 client_data.client_number)

        for port_data in client_data.ports:
            midi_port = MidiPort(port_data.port_name,
                                 port_data.port_number,
                                 port_data.port_index,
                                 port_data.system_port_name,
                                 direction=(MidiPort.OUT
                                            if port_data.direction == MidiPort.OUT  # noqa
                                            else MidiPort.IN),
                                 midi_out=midi_out,
                                 midi_in=midi_in)
            if port_data.direction == MidiPort.OUT:
                midi_client.append_out_port(midi_port)
            elif port_data.direction == MidiPort.IN:
                midi_client.append_in_port(midi_port)

        found_launchpads.append(LaunchpadMiniMk3(midi_client))

    return found_launchpads



import logging
from logging import FileHandler
import os
import sys

LOGFILE = os.environ.get('LOGFILE')
FORMATTER = logging.Formatter('[%(levelname)s] %(name)s (%(asctime)s): %(message)s',  # noqa
                              '%m-%d-%Y %I:%M:%S')
LOG_LEVELS = {
    'debug': logging.DEBUG,
    'info': logging.INFO,
    'warning': logging.WARNING,
    'error': logging.ERROR,
    'critical': logging.CRITICAL
}


def get_console_handler():
    consoleHandler = logging.StreamHandler(sys.stdout)
    consoleHandler.setFormatter(FORMATTER)
    return consoleHandler


def get_file_handler():
    if LOGFILE:
        if not os.path.exists(os.path.dirname(LOGFILE)):
            os.makedirs(os.path.dirname(LOGFILE))
        fileHandler = FileHandler(LOGFILE)
        fileHandler.setFormatter(FORMATTER)
        return fileHandler
    return None


def getLogger(name):
    logger = logging.getLogger(name)
    logger.setLevel(LOG_LEVELS[os.environ.get('LOGLEVEL', 'warning').lower()])
    logger.addHandler(get_console_handler())
    file_handler = get_file_handler()
    if file_handler:
        logger.addHandler(file_handler)
    logger.propagate = False
    return logger



"""
MIDI messages for the Launchpad Mini MK3.
"""

from abc import ABC
from functools import reduce


class SysExMessages:
    DEVICE_INQUIRY = [0xf0, 0x7e, 0xe7, 0xf0, 0x06, 0x01, 0xf7]

    class Interfaces:
        DAW = [0xf0, 0x00, 0x20, 0x29, 0x02, 0x0d, 0x10, 0x00, 0xf7]
        MIDI = [0xf0, 0x00, 0x20, 0x29, 0x02, 0x0d, 0x10, 0x01, 0xf7]
        READBACK = [0xf0, 0x00, 0x20, 0x29, 0x02, 0x0d, 0x10, 0xf7]

    class Layouts:
        SESSION = [0xf0, 0x00, 0x20, 0x29, 0x02, 0x0d, 0x00, 0x00, 0xf7]
        CUSTOM_1 = [0xf0, 0x00, 0x20, 0x29, 0x02, 0x0d, 0x00, 0x04, 0xf7]
        CUSTOM_2 = [0xf0, 0x00, 0x20, 0x29, 0x02, 0x0d, 0x00, 0x05, 0xf7]
        CUSTOM_3 = [0xf0, 0x00, 0x20, 0x29, 0x02, 0x0d, 0x00, 0x06, 0xf7]
        DAW_FADERS = [0xf0, 0x00, 0x20, 0x29, 0x02, 0x0d, 0x00, 0x0d, 0xf7]
        PROG = [0xf0, 0x00, 0x20, 0x29, 0x02, 0x0d, 0x00, 0x7f, 0xf7]
        READBACK = [0xf0, 0x00, 0x20, 0x29, 0x02, 0x0d, 0x00, 0xf7]

    class Modes:
        LIVE = [0xf0, 0x00, 0x20, 0x29, 0x02, 0x0d, 0x0e, 0x00, 0xf7]
        PROG = [0xf0, 0x00, 0x20, 0x29, 0x02, 0x0d, 0x0e, 0x01, 0xf7]
        READBACK = [0xf0, 0x00, 0x20, 0x29, 0x02, 0x0d, 0x0e, 0xf7]


class MidiMessage(ABC):
    @property
    def data(self):
        return []


class Constants:
    DEFAULT_COLOR_ID = 9
    MIDI_MIN_VALUE = 0
    MIDI_MAX_VALUE = 0x7f

    class LightingMode:
        OFF = 0x80
        STATIC = 0x90
        FLASH = 0x91

    class LightingType:
        STATIC = 0x00
        FLASH = 0x01
        PULSE = 0x02
        RGB = 0x03

    class MidiWord:
        NOTE_HEADER = 0x90
        CC_HEADER = 0xb0


class ColorspecFragment:
    def __init__(self, lighting_type, led_index, *lighting_data):
        self._lighting_type = lighting_type
        self._led_index = led_index
        self._lighting_data = list(lighting_data)

    @property
    def data(self):
        return ([self._lighting_type,
                self._led_index]
                + self._lighting_data)

    def append(self, lighting_data):
        self._lighting_data.extend(lighting_data)

    def __repr__(self):
        return ('ColorspecFragment('
                f'lighting_type={self._lighting_type}, '
                f'led_index={self._led_index}, '
                f'lighting_data={self._lighting_data})')

    def __str__(self):
        fragment_str = str([self._lighting_type, self._led_index]
                           + list(self._lighting_data))
        return fragment_str


class Colorspec(MidiMessage):
    def __init__(self, *fragments):
        self._start_clause = [0xf0, 0x00, 0x20, 0x29, 0x02, 0x0d, 0x03]
        self._fragments = list(fragments)
        self._end_clause = [0xf7]

    @MidiMessage.data.getter
    def data(self):
        payload = list(reduce(lambda data, fragment: data + fragment.data,
                              self._fragments, []))  # noqa
        return (self._start_clause
                + payload
                + self._end_clause)

    def append(self, fragment):
        if not isinstance(fragment, ColorspecFragment):
            raise ValueError('Must be a ColorspecFragment.')
        self._fragments.append(fragment)

    def __repr__(self):
        return f'Colorspec({self.data})'


class Lighting(MidiMessage):
    def __init__(self, lighting_mode, midi_value, color_id):
        self._lighting_mode = lighting_mode
        self._midi_value = midi_value
        self._color_id = color_id

    def __repr__(self):
        return ('Lighting('
                f'lighting_mode={self._lighting_mode}, '
                f'midi_value={self._midi_value}, '
                f'color_id={self.color_id})')

    @MidiMessage.data.getter
    def data(self):
        return [self._lighting_mode,
                self._midi_value,
                self._color_id]

    @property
    def lighting_mode(self):
        return self._lighting_mode

    @property
    def midi_value(self):
        return self._midi_value

    @property
    def color_id(self):
        return self._color_id



import time
from ..midi_messages import Colorspec,\
                            ColorspecFragment,\
                            Constants


class RawBitmapRenderer:
    def __init__(self,
                 raw_bitmap,
                 matrix, *,
                 angle=0,
                 flip_axis='',
                 fg_color=None,
                 bg_color=None):
        self._raw_bitmap = raw_bitmap
        self._matrix = matrix
        self._fg_color = fg_color
        self._bg_color = bg_color
        self._angle = angle
        self._flip_axis = flip_axis

    def render(self):
        colorspec_fragments = []
        for led, bit in zip(self._matrix.led_range(rotation=self._angle,
                                                   flip_axis=self._flip_axis),
                            self._raw_bitmap):
            bit_config = self._raw_bitmap.config[led.name]
            lighting_type = bit_config.lighting_type
            led_index = led.midi_value
            lighting_data = self._determine_lighting_data(bit_config,
                                                          bit,
                                                          self._fg_color,
                                                          self._bg_color)
            fragment = ColorspecFragment(lighting_type,
                                         led_index,
                                         *lighting_data)
            colorspec_fragments.append(fragment)
        payload = Colorspec(*colorspec_fragments)
        self._matrix.launchpad.send_message(payload)

    def _determine_lighting_data(self, config, bit, fg_color, bg_color):
        lighting_data = []
        if config.lighting_type == Constants.LightingType.FLASH:
            lighting_data = ([config.lighting_data.on_state.color.id]
                             if bit
                             else [config.lighting_data.off_state.color.id])
        elif config.lighting_type == Constants.LightingType.PULSE:
            lighting_data = ([config.lighting_data.on_state.color.id]
                             if bit
                             else [config.lighting_data.off_state.color.id])
        elif config.lighting_type == Constants.LightingType.RGB:
            lighting_data = (config.lighting_data.on_state.color.rgb
                             if bit
                             else config.lighting_data.off_state.color.rgb)
        else:
            fg_color_id = fg_color.color_id if fg_color else 0
            bg_color_id = bg_color.color_id if bg_color else 0
            on_state = ([config.lighting_data.on_state.color.id]
                        if config.lighting_data
                        else [fg_color_id])
            off_state = ([config.lighting_data.off_state.color.id]
                         if config.lighting_data
                         else [bg_color_id])
            lighting_data = (on_state
                             if bit
                             else off_state)
        return lighting_data


class TextRenderer:
    def __init__(self,
                 text,
                 string,
                 period,
                 direction,
                 matrix, *,
                 timeout,
                 count,
                 cycle_func):
        self._text = text
        self._string = string
        self._period = period
        self._direction = direction
        self._matrix = matrix
        self._timeout = timeout
        self._count = count
        self._cycle_func = cycle_func

    def render(self):
        time_left = self._timeout
        rotations_left = self._count
        text_width = len(self._text) * self._matrix.width
        while time_left and rotations_left:
            for _ in range(text_width):
                if self._direction == 'right':
                    self._string.shift_right()
                else:
                    self._string.shift_left()
                CharacterRenderer(self._string.character_to_render,
                                  self._matrix,
                                  angle=self._string.angle,
                                  flip_axis=self._string.flip_axis).render()
                if self._cycle_func:
                    self._cycle_func(_/text_width, self._matrix.launchpad)
                time.sleep(self._period)
                time_left = (max(time_left - self._period, 0)
                             if time_left != -1
                             else time_left)
            rotations_left = (max(rotations_left - 1, 0)
                              if rotations_left != -1
                              else rotations_left)


class CharacterRenderer:
    def __init__(self,
                 character,
                 matrix, *,
                 angle=0,
                 flip_axis=''):
        self._renderer = RawBitmapRenderer(character.raw_bitmap,
                                           matrix,
                                           angle=angle,
                                           flip_axis=flip_axis,
                                           fg_color=character.fg_color,
                                           bg_color=character.bg_color)

    def render(self):
        self._renderer.render()


class BitmapRenderer:
    def __init__(self,
                 raw_bitmap,
                 matrix, *,
                 fg_color,
                 bg_color):
        self._renderer = RawBitmapRenderer(raw_bitmap,
                                           matrix,
                                           fg_color=fg_color,
                                           bg_color=bg_color)

    def render(self):
        self._renderer.render()


class FrameRenderer:
    def __init__(self,
                 raw_bitmap,
                 matrix):
        self._renderer = RawBitmapRenderer(raw_bitmap,
                                           matrix)

    def render(self):
        self._renderer.render()


class MovieRenderer:
    def __init__(self,
                 raw_bitmaps,
                 framerate,
                 matrix, *,
                 count=None):
        self._raw_bitmaps = raw_bitmaps
        self._framerate = max(1, min(60, framerate))
        self._matrix = matrix
        self._angle = 0
        self._flip_axis = 'x'
        self._count = -1 if not count else count

    def render(self):
        period = 1 / self._framerate
        rotations_left = self._count
        while rotations_left:
            for bitmap in self._raw_bitmaps:
                renderer = RawBitmapRenderer(bitmap,
                                             self._matrix)
                renderer.render()
                time.sleep(period)
            rotations_left = (max(rotations_left - 1, 0)
                              if rotations_left != -1
                              else rotations_left)



import re


class RgbColor:
    """
    An RGB color.
    """
    _RGB_LENGTH = 3
    _MIN_HEX_LENGTH = 4
    _MAX_HEX_LENGTH = 7
    _MIN_COLOR_VALUE = 0
    _MAX_COLOR_VALUE = 255

    def __init__(self, value):
        if (not isinstance(value, str)
                and not isinstance(value, (tuple, list))):
            raise TypeError('RGB color must be str or tuple or list.')
        if (isinstance(value, str)
            and len(value) != RgbColor._MIN_HEX_LENGTH
                and len(value) != RgbColor._MAX_HEX_LENGTH):
            raise ValueError('Only hex values in the '
                             '#rgb and #rrggbb formats are accepted.')
        self._parse(value)

    def __repr__(self):
        return f'RgbColor(r={self.r}, g={self.g}, b={self.b})'

    @staticmethod
    def is_valid(value):
        """
        Returns `True` if `value` is a valid (r,g,b) or hex value,
        otherwise returns `False`.
        """
        if isinstance(value, str):
            match = re.match('^#(?:[0-9a-fA-F]{3}){1,2}$', value)
            return True if match else False
        elif isinstance(value, (tuple, list)):
            valid_values = list(filter(lambda rgb_value: min(RgbColor._MAX_COLOR_VALUE,  # noqa
                                                             max(rgb_value, RgbColor._MIN_COLOR_VALUE)) == rgb_value,  # noqa
                                       value))
            return (len(valid_values) == len(value)
                    and len(value) == RgbColor._RGB_LENGTH)
        return False

    @property
    def value(self):
        """
        RGB value, as string.
        """
        return self._value

    @property
    def r(self):
        """
        Red value
        """
        return self._r >> 1

    @property
    def g(self):
        """
        Green value
        """
        return self._g >> 1

    @property
    def b(self):
        """
        Blue value
        """
        return self._b >> 1

    def _parse(self, value):
        if (isinstance(value, str)
                and len(value) == RgbColor._MIN_HEX_LENGTH):
            self._r = int(value[1], 16)
            self._g = int(value[2], 16)
            self._b = int(value[3], 16)
            self._value = value
        elif (isinstance(value, str)
                and len(value) == RgbColor._MAX_HEX_LENGTH):
            self._r = int(value[1:3], 16)
            self._g = int(value[3:5], 16)
            self._b = int(value[5:7], 16)
            self._value = value
        elif (isinstance(value, (tuple, list))
                and len(value) == RgbColor._RGB_LENGTH):
            self._r = value[0]
            self._g = value[1]
            self._b = value[2]
            self._value = value


class ColorShade:
    """
    A color shade.

    A color shade is color from the Launchpad's color palette.
    There are 128 valid color shades, ranging from ID 0 to 127
    (with ID 0 used for OFF).
    """

    MIN_COLOR_ID = 0
    MAX_COLOR_ID = 127
    HEX_VALUES = [
            '#616161',
            '#b3b3b3',
            '#ffffff',
            '#ffb3b3',
            '#ff6161',  # 5
            '#dd6161',
            '#b36161',
            '#fff3d5',
            '#ffb361',
            '#dd8c61',  # 10
            '#b37661',
            '#ffeea1',
            '#ffff61',
            '#dddd61',
            '#b3b361',  # 15
            '#ddffa1',
            '#a1dd61',
            '#81b361',
            '#c2ffb3',
            '#c2ffb3',  # 20
            '#61ff61',
            '#61dd61',
            '#61b361',
            '#c2ffc2',
            '#61ff8c',  # 25
            '#61dd76',
            '#61b36b',
            '#c2ffcc',
            '#61ffcc',
            '#61dda1',  # 30
    ]

    def __init__(self, color_id, color_group):
        self._color_id = color_id
        self._color_group = color_group

    def __repr__(self):
        return ('ColorShade('
                f'group={self.color_group}, '
                f'id={self.color_id})')

    @staticmethod
    def is_valid_id(value):
        """
        Returns `True` if `value` is a valid color ID,
        otherwise returns `False`.
        """
        return (True
                if isinstance(value, int)
                and value >= ColorShade.MIN_COLOR_ID
                and value <= ColorShade.MAX_COLOR_ID
                else False)

    @property
    def hex(self):
        """Hex color code."""
        return ColorShade.HEX_VALUES[self._color_id]

    @property
    def color_id(self):
        """Unique color ID."""
        return self._color_id

    @property
    def color_group(self):
        """Color group."""
        return self._color_group


class ColorPalette:
    """
    Color palette of Launchpad.

    The color palette contains 127 predefined color shades.
    """
    class Red:
        SHADE_1 = ColorShade(0x04, 'red')
        SHADE_2 = ColorShade(0x05, 'red')
        SHADE_3 = ColorShade(0x06, 'red')
        SHADE_4 = ColorShade(0x07, 'red')
        SHADE_5 = ColorShade(0x48, 'red')
        SHADE_6 = ColorShade(0x6a, 'red')
        SHADE_7 = ColorShade(0x6b, 'red')
        SHADE_8 = ColorShade(0x78, 'red')
        SHADE_9 = ColorShade(0x79, 'red')

    class Orange:
        SHADE_1 = ColorShade(0x09, 'orange')
        SHADE_2 = ColorShade(0x0a, 'orange')
        SHADE_3 = ColorShade(0x0b, 'orange')
        SHADE_4 = ColorShade(0x3c, 'orange')
        SHADE_5 = ColorShade(0x53, 'orange')
        SHADE_6 = ColorShade(0x54, 'orange')
        SHADE_7 = ColorShade(0x60, 'orange')
        SHADE_8 = ColorShade(0x6c, 'orange')
        SHADE_9 = ColorShade(0x7e, 'orange')
        SHADE_10 = ColorShade(0x7f, 'orange')

    class Yellow:
        SHADE_1 = ColorShade(0x08, 'yellow')
        SHADE_2 = ColorShade(0x0c, 'yellow')
        SHADE_3 = ColorShade(0x3d, 'yellow')
        SHADE_4 = ColorShade(0x3e, 'yellow')
        SHADE_5 = ColorShade(0x49, 'yellow')
        SHADE_6 = ColorShade(0x4a, 'yellow')
        SHADE_7 = ColorShade(0x61, 'yellow')
        SHADE_8 = ColorShade(0x63, 'yellow')
        SHADE_9 = ColorShade(0x64, 'yellow')
        SHADE_10 = ColorShade(0x69, 'yellow')
        SHADE_11 = ColorShade(0x6d, 'yellow')
        SHADE_12 = ColorShade(0x71, 'yellow')
        SHADE_13 = ColorShade(0x7c, 'yellow')
        SHADE_14 = ColorShade(0x7d, 'yellow')

    class Blue:
        SHADE_1 = ColorShade(0x24, 'blue')
        SHADE_2 = ColorShade(0x25, 'blue')
        SHADE_3 = ColorShade(0x26, 'blue')
        SHADE_4 = ColorShade(0x27, 'blue')
        SHADE_5 = ColorShade(0x28, 'blue')
        SHADE_6 = ColorShade(0x29, 'blue')
        SHADE_7 = ColorShade(0x2a, 'blue')
        SHADE_8 = ColorShade(0x2b, 'blue')
        SHADE_9 = ColorShade(0x41, 'blue')
        SHADE_10 = ColorShade(0x42, 'blue')
        SHADE_11 = ColorShade(0x4e, 'blue')
        SHADE_12 = ColorShade(0x4f, 'blue')
        SHADE_13 = ColorShade(0x5a, 'blue')
        SHADE_14 = ColorShade(0x5b, 'blue')
        SHADE_15 = ColorShade(0x5c, 'blue')
        SHADE_16 = ColorShade(0x67, 'blue')
        SHADE_17 = ColorShade(0x68, 'blue')

    class Green:
        SHADE_1 = ColorShade(0x0d, 'green')
        SHADE_2 = ColorShade(0x0e, 'green')
        SHADE_3 = ColorShade(0x0f, 'green')
        SHADE_4 = ColorShade(0x10, 'green')
        SHADE_5 = ColorShade(0x11, 'green')
        SHADE_6 = ColorShade(0x12, 'green')
        SHADE_7 = ColorShade(0x13, 'green')
        SHADE_8 = ColorShade(0x14, 'green')
        SHADE_9 = ColorShade(0x15, 'green')
        SHADE_10 = ColorShade(0x16, 'green')
        SHADE_11 = ColorShade(0x17, 'green')
        SHADE_12 = ColorShade(0x18, 'green')
        SHADE_13 = ColorShade(0x19, 'green')
        SHADE_14 = ColorShade(0x1a, 'green')
        SHADE_15 = ColorShade(0x1b, 'green')
        SHADE_16 = ColorShade(0x1c, 'green')
        SHADE_17 = ColorShade(0x1d, 'green')
        SHADE_18 = ColorShade(0x1e, 'green')
        SHADE_19 = ColorShade(0x1f, 'green')
        SHADE_20 = ColorShade(0x20, 'green')
        SHADE_21 = ColorShade(0x21, 'green')
        SHADE_22 = ColorShade(0x22, 'green')
        SHADE_23 = ColorShade(0x23, 'green')
        SHADE_24 = ColorShade(0x3f, 'green')
        SHADE_25 = ColorShade(0x40, 'green')
        SHADE_26 = ColorShade(0x41, 'green')
        SHADE_27 = ColorShade(0x44, 'green')
        SHADE_28 = ColorShade(0x4b, 'green')
        SHADE_29 = ColorShade(0x4c, 'green')
        SHADE_30 = ColorShade(0x4d, 'green')
        SHADE_31 = ColorShade(0x55, 'green')
        SHADE_32 = ColorShade(0x56, 'green')
        SHADE_33 = ColorShade(0x57, 'green')
        SHADE_34 = ColorShade(0x58, 'green')
        SHADE_35 = ColorShade(0x59, 'green')
        SHADE_36 = ColorShade(0x62, 'green')
        SHADE_37 = ColorShade(0x65, 'green')
        SHADE_38 = ColorShade(0x66, 'green')
        SHADE_39 = ColorShade(0x6e, 'green')
        SHADE_40 = ColorShade(0x6f, 'green')
        SHADE_41 = ColorShade(0x72, 'green')
        SHADE_42 = ColorShade(0x7a, 'green')
        SHADE_43 = ColorShade(0x7b, 'green')

    class Violet:
        SHADE_1 = ColorShade(0x2c, 'violet')
        SHADE_2 = ColorShade(0x2d, 'violet')
        SHADE_3 = ColorShade(0x2e, 'violet')
        SHADE_4 = ColorShade(0x2f, 'violet')
        SHADE_5 = ColorShade(0x30, 'violet')
        SHADE_6 = ColorShade(0x31, 'violet')
        SHADE_7 = ColorShade(0x32, 'violet')
        SHADE_8 = ColorShade(0x33, 'violet')
        SHADE_9 = ColorShade(0x34, 'violet')
        SHADE_10 = ColorShade(0x35, 'violet')
        SHADE_11 = ColorShade(0x36, 'violet')
        SHADE_12 = ColorShade(0x37, 'violet')
        SHADE_13 = ColorShade(0x38, 'violet')
        SHADE_14 = ColorShade(0x39, 'violet')
        SHADE_15 = ColorShade(0x3a, 'violet')
        SHADE_16 = ColorShade(0x3b, 'violet')
        SHADE_17 = ColorShade(0x43, 'violet')
        SHADE_18 = ColorShade(0x45, 'violet')
        SHADE_19 = ColorShade(0x46, 'violet')
        SHADE_20 = ColorShade(0x47, 'violet')
        SHADE_21 = ColorShade(0x50, 'violet')
        SHADE_22 = ColorShade(0x51, 'violet')
        SHADE_23 = ColorShade(0x52, 'violet')
        SHADE_24 = ColorShade(0x5d, 'violet')
        SHADE_25 = ColorShade(0x5e, 'violet')
        SHADE_26 = ColorShade(0x5f, 'violet')
        SHADE_27 = ColorShade(0x70, 'violet')
        SHADE_28 = ColorShade(0x73, 'violet')
        SHADE_29 = ColorShade(0x74, 'violet')

    class White:
        SHADE_1 = ColorShade(0x01, 'white')
        SHADE_2 = ColorShade(0x02, 'white')
        SHADE_3 = ColorShade(0x03, 'white')
        SHADE_4 = ColorShade(0x75, 'white')
        SHADE_5 = ColorShade(0x76, 'white')
        SHADE_6 = ColorShade(0x77, 'white')


class ColorShadeStore:
    """
    Color shade store for the Launchpad.
    """
    COLOR_GROUPS = ['red', 'orange', 'yellow', 'green',
                    'blue', 'violet', 'white']
    COLOR_GROUP_SYMBOLS = list(map(lambda color: color[0],
                                   COLOR_GROUPS))

    def __iter__(self):
        for color_group in ColorShadeStore.COLOR_GROUPS:
            color_shades_in_group = ColorPalette.__dict__[color_group.capitalize()].__dict__  # noqa
            for _, color_shade in color_shades_in_group.items():
                if isinstance(color_shade, ColorShade):
                    yield color_shade

    def contains(self, value):
        """
        Searches for color shade `value` and returns `True` if
        the color shade exists, otherwise returns `False`.
        """
        color, _ = self._parse(value)
        return True if color else False

    def find(self, value):
        """
        Searches for color shade `value` and returns the
        `ColorShade` if found, otherwise returns `None`.
        """
        if isinstance(value, ColorShade):
            return value
        return self._find_shade(*self._parse(value))

    def _find_shade(self, color, color_shade_id=1):
        if not color:
            return None

        color_shade = None
        color_shades_in_group = ColorPalette.__dict__[color.capitalize()].__dict__  # noqa
        if color_shade_id:
            shade_clause = f'SHADE_{color_shade_id}'
            if shade_clause in color_shades_in_group:
                color_shade = color_shades_in_group[shade_clause]
        else:
            color_shade = color_shades_in_group['SHADE_1']

        return color_shade

    def _color_from_symbol(self, symbol):
        return ColorShadeStore.COLOR_GROUPS[ColorShadeStore.COLOR_GROUP_SYMBOLS.index(symbol)]  # noqa

    # Valid input: red, o, y2, green3
    def _parse(self, value):
        color = None
        color_shade_id = None

        if isinstance(value, str):
            has_digit_suffix = re.search('[0-9]+', value)
            if has_digit_suffix:
                match = re.match('([a-z]+)([0-9]+)', value.lower())
                if match and len(match.groups()) == 2:
                    color = (match.group(1)
                             if match.group(1) in ColorShadeStore.COLOR_GROUPS
                             else None)
                    color = (self._color_from_symbol(match.group(1))
                             if not color and match.group(1)
                             in ColorShadeStore.COLOR_GROUP_SYMBOLS
                             else color)
                    color_shade_id = int(match.group(2))
                else:
                    color = None
                    color_shade_id = None
            elif value.lower() in ColorShadeStore.COLOR_GROUPS:
                color = value.lower()
                color_shade_id = 1
            elif value.lower() in ColorShadeStore.COLOR_GROUP_SYMBOLS:
                color = self._color_from_symbol(value.lower())
                color_shade_id = 1

        return (color, color_shade_id)



"""
Utility classes for Launchpad Mini MK3.
"""

import enum
import time
import re
import platform
from collections import namedtuple
from . import logging
from ..match import Match

logger = logging.getLogger(__name__)

MIDI_MESSAGE_LENGTH = 9


class MidiEvent:
    """
    A MIDI event.

    A MIDI event is received every time the Launchpad's MIDI port
    is read.
    """

    def __init__(self, message, deltatime=0):
        self._message = message
        self._deltatime = deltatime

    def __eq__(self, other):
        if not isinstance(other, (MidiEvent, list)):
            return False
        if isinstance(other, list):
            return self.message == other
        return self.message == other.message

    @property
    def message(self):
        return self._message

    @property
    def deltatime(self):
        return self._deltatime

    def __repr__(self):
        return ('MidiEvent('
                f'message={self.message}, '
                f'deltatime={self.deltatime})')


class ButtonEvent:
    """
    A button event.

    A button event is received every time a button on the Launchpad
    is pushed.
    """

    PRESS = 'press'
    RELEASE = 'release'
    PRESS_RELEASE = 'press_release'

    def __init__(self, midi_event, buttons):
        self._midi_event = midi_event
        self._button = self._determine_button(midi_event, buttons)

    def __eq__(self, other):
        if not isinstance(other, ButtonEvent):
            return False
        return self.midi_event == other.midi_event

    def __repr__(self):
        return ("ButtonEvent("
                f"button='{self.button.name}', "
                f"type='{self.type}', "
                f"deltatime={self.deltatime})")

    @property
    def message(self):
        return self._midi_event.message

    @property
    def deltatime(self):
        return self._midi_event.deltatime

    @property
    def button(self):
        return self._button

    @property
    def type(self):
        if not self._midi_event:
            return ''
        return (ButtonEvent.RELEASE
                if self._midi_event
                and len(self._midi_event.message) == 3
                and self._midi_event.message[2] == 0x0
                else ButtonEvent.PRESS)

    def _determine_button(self, midi_event, buttons):
        found_buttons = (list(filter(lambda b: b.midi_value == midi_event.message[1], buttons))  # noqa
                         if midi_event and midi_event.message and len(midi_event.message) == 3  # noqa
                         else None)
        return (found_buttons[0]
                if found_buttons and len(found_buttons)
                else None)


class MidiPort:
    """
    A MIDI port.
    """

    OUT = 'out'
    IN = 'in'
    DEFAULT_CLIENT_NAME = 'lpminimk3'

    def __init__(self, port_name, port_number, port_index,
                 system_port_name, *, direction,
                 midi_in=None, midi_out=None,
                 virtual=False):
        self._port_name = port_name
        self._port_number = port_number
        self._port_index = port_index
        self._system_port_name = system_port_name
        self._midi_in = midi_in
        self._midi_out = midi_out
        self._direction = direction
        self._virtual = virtual
        if midi_in:
            midi_in.ignore_types(sysex=False, timing=False)

    def __eq__(self, other):
        if not isinstance(other, MidiPort):
            return False
        return self.system_port_name == other.system_port_name

    def __repr__(self):
        return ("MidiPort("
                f"name='{self.port_name}', "
                f"number={self.port_number}, "
                f"index={self.port_index})")

    def __exit__(self, *args, **kwargs):
        self.close()

    @property
    def port_name(self):
        return self._port_name

    @property
    def port_number(self):
        return self._port_number

    @property
    def port_index(self):
        return self._port_index

    @property
    def system_port_name(self):
        return self._system_port_name

    @property
    def midi_in_handle(self):
        return self._midi_in

    @property
    def midi_out_handle(self):
        return self._midi_out

    def is_open(self):
        if self._midi_in and self._direction == MidiPort.IN:
            return self._midi_in.is_port_open()
        elif self._midi_out and self._direction == MidiPort.OUT:
            return self._midi_out.is_port_open()
        return False

    def open(self):
        if (self._direction == MidiPort.OUT
                and not self._midi_out.is_port_open()):
            if self._virtual:
                self._midi_out.open_virtual_port(self._system_port_name)
            else:
                self._midi_out.open_port(self.port_index, MidiPort.OUT)

            if platform.system() != 'Windows' and platform.system() != 'Darwin':
                self._midi_out.set_client_name(MidiPort.DEFAULT_CLIENT_NAME)

        elif (self._direction == MidiPort.IN
                and not self._midi_in.is_port_open()):
            if self._virtual:
                self._midi_in.open_virtual_port(self._system_port_name)
            else:
                self._midi_in.open_port(self.port_index, MidiPort.IN)

            if platform.system() != 'Windows' and platform.system() != 'Darwin':
                self._midi_in.set_client_name(MidiPort.DEFAULT_CLIENT_NAME)

    def close(self):
        if self.is_open():
            if self._midi_out and self._direction == MidiPort.OUT:
                self._midi_out.close_port()
            elif self._midi_in and self._direction == MidiPort.IN:
                self._midi_in.close_port()

    def send_message(self, message):
        if (not message
                or (not isinstance(message, list)
                    and not hasattr(message, 'data'))):
            raise TypeError('Message must be of type list or MidiMessage.')
        message = (message.data
                   if hasattr(message, 'data')
                   else message)

        if self._midi_in and self._direction == MidiPort.IN:
            self._midi_in.send_message(message)
            logger.debug(f'MIDI message sent: {message}')
        elif self._midi_out and self._direction == MidiPort.OUT:
            self._midi_out.send_message(message)
            logger.debug(f'MIDI message sent: {message}')
        else:
            raise RuntimeError('Failed to send message.')

    def poll_for_event(self, *, timeout=5, match=None, read_delay=.001):
        assert self._midi_in
        event = None
        polling = True
        elapsed = 0 if timeout else -1
        timeout = 0 if not timeout else timeout
        timeout = 0 if timeout < 0 else timeout
        while polling and timeout > elapsed:
            try:
                raw_message = self._midi_in.get_message()
                event = MidiEvent(*raw_message) if raw_message else None
                logger.debug(f'MIDI event: {event}')
                if event and not match:
                    polling = False
                elif event and isinstance(match, list) and match == event.message:  # noqa
                    polling = False
                elif event and isinstance(match, Match) and match.contains(event.message):  # noqa
                    polling = False
                time.sleep(read_delay)
                elapsed += .1 if timeout and timeout > 0 else 0
            except KeyboardInterrupt:
                logger.debug('\nPolling terminated.')
                raise
        return event

    def clear_event_queue(self, *, read_delay=.001):
        while self._midi_in and self._midi_in.get_message():
            time.sleep(read_delay)


class Interface:
    """
    Interface of the launchpad.
    """

    DAW = 'daw'
    MIDI = 'midi'
    READBACK_POSITION = 7

    class MidiWord(enum.IntEnum):
        DAW = 0x00
        MIDI = 0x01

    def __init__(self, midi_event):
        if len(midi_event.message) != MIDI_MESSAGE_LENGTH:
            raise RuntimeError('Unexpected MIDI message length; '
                               f'expected {MIDI_MESSAGE_LENGTH}, '
                               f'got {len(midi_event.message)}.')
        self._midi_event = midi_event
        midi_value = midi_event.message[Interface.READBACK_POSITION]
        self._interface = (Interface.MIDI
                           if midi_value == Interface.MidiWord.MIDI
                           else Interface.DAW)

    def __eq__(self, other):
        if (not isinstance(other, str)
                or (other.lower() != Interface.MIDI
                    and other.lower() != Interface.DAW)):
            return False
        return self.midi_event == other.midi_event

    def __repr__(self):
        return ('Interface()'
                if not self._interface
                else f"Interface('Interface.{self._interface.upper()}')")

    @property
    def midi_event(self):
        return self._midi_event


class Mode:
    """
    A Launchpad mode.
    """

    LIVE = 'live'
    PROG = 'prog'
    READBACK_POSITION = 7

    class MidiWord(enum.IntEnum):
        LIVE = 0x00
        PROG = 0x01

    def __init__(self, midi_event):
        if len(midi_event.message) != MIDI_MESSAGE_LENGTH:
            raise RuntimeError('Unexpected MIDI message length; '
                               f'expected {MIDI_MESSAGE_LENGTH}, '
                               f'got {midi_event.message}.')
        self._midi_event = midi_event
        midi_value = midi_event.message[Mode.READBACK_POSITION]
        self._mode = (Mode.LIVE
                      if midi_value == Mode.MidiWord.LIVE
                      else Mode.PROG)

    def __eq__(self, other):
        if (not isinstance(other, str)
                or (other.lower() != Mode.LIVE
                    and other.lower() != Mode.PROG)):
            return False
        return self.midi_event == other.midi_event

    def __repr__(self):
        return ('Mode()'
                if not self._mode
                else f"Mode('Mode.{self._mode.upper()}')")

    @property
    def midi_event(self):
        return self._midi_event


class Layout:
    """
    A Launchpad layout.
    """

    SESSION = 'session'
    CUSTOM_1 = 'custom_1'
    CUSTOM_2 = 'custom_2'
    CUSTOM_3 = 'custom_3'
    DAW_FADERS = 'daw_faders'
    PROG = 'prog'
    READBACK_POSITION = 7

    class MidiWord(enum.IntEnum):
        SESSION = 0x00
        CUSTOM_1 = 0x04
        CUSTOM_2 = 0x05
        CUSTOM_3 = 0x06
        DAW_FADERS = 0x0d
        PROG = 0x7f

    def __init__(self, midi_event):
        if len(midi_event.message) != MIDI_MESSAGE_LENGTH:
            raise RuntimeError('Unexpected MIDI message length; '
                               f'expected {MIDI_MESSAGE_LENGTH}, '
                               f'got {len(midi_event.message)}.')
        self._midi_event = midi_event
        midi_value = midi_event.message[Layout.READBACK_POSITION]
        self._layout = self._determine_layout(midi_value)

    def __eq__(self, other):
        if (not isinstance(other, str)
                or (other.lower() != Layout.SESSION
                    and other.lower() != Layout.CUSTOM_1
                    and other.lower() != Layout.CUSTOM_2
                    and other.lower() != Layout.CUSTOM_3
                    and other.lower() != Layout.DAW_FADERS
                    and other.lower() != Layout.PROG)):
            return False
        return self.midi_event == other.midi_event

    def __repr__(self):
        return ('Layout()'
                if not self._layout
                else f"Layout('Layout.{self._layout.upper()}')")

    @property
    def midi_event(self):
        return self._midi_event

    def _determine_layout(self, midi_value):
        if midi_value == Layout.MidiWord.SESSION:
            layout = Layout.SESSION
        elif midi_value == Layout.MidiWord.CUSTOM_1:
            layout = Layout.CUSTOM_1
        elif midi_value == Layout.MidiWord.CUSTOM_2:
            layout = Layout.CUSTOM_2
        elif midi_value == Layout.MidiWord.CUSTOM_3:
            layout = Layout.CUSTOM_3
        elif midi_value == Layout.MidiWord.DAW_FADERS:
            layout = Layout.DAW_FADERS
        else:
            layout = Layout.PROG
        return layout


class MidiClient:
    """
    A MIDI client.
    """

    def __init__(self, client_name, client_number):
        self._client_name = client_name
        self._client_number = client_number
        self._out_ports = []
        self._in_ports = []

    def __eq__(self, other):
        if not isinstance(other, MidiClient):
            return False
        return (self.client_name == other.client_name
                and self.client_number == other.client_number)

    def __repr__(self):
        return ("MidiClient("
                f"name='{self.client_name}', "
                f"number={self.client_number})")

    @property
    def daw_out_port(self):
        daw_ports = list(filter(lambda port: re.search(r'1|DA',
                                port.port_name),
                                self._out_ports))
        return daw_ports[0] if len(daw_ports) > 0 else None

    @property
    def daw_in_port(self):
        daw_ports = list(filter(lambda port: re.search(r'1|DA',
                                port.port_name),
                                self._in_ports))
        return daw_ports[0] if len(daw_ports) > 0 else None

    @property
    def midi_out_port(self):
        midi_ports = list(filter(lambda port: re.search(r'2|MI',
                                 port.port_name),
                                 self._out_ports))
        return midi_ports[0] if len(midi_ports) > 0 else None

    @property
    def midi_in_port(self):
        midi_ports = list(filter(lambda port: re.search(r'2|MI',
                                 port.port_name),
                                 self._in_ports))
        return midi_ports[0] if len(midi_ports) > 0 else None

    @property
    def ports(self):
        return self._out_ports + self._in_ports

    @property
    def out_ports(self):
        return self._out_ports

    @property
    def in_ports(self):
        return self._in_ports

    @property
    def client_name(self):
        return self._client_name

    @property
    def client_number(self):
        return self._client_number

    def is_open(self):
        return all(port.is_open() for port in self.ports)

    def open(self, interface=Interface.MIDI):
        self.close()
        if interface == Interface.DAW:
            self.daw_in_port.open()
            self.daw_out_port.open()
        elif interface == Interface.MIDI:
            self.midi_in_port.open()
            self.midi_out_port.open()
        else:
            raise ValueError('Must be a valid Interface.')

    def close(self):
        for port in self.ports:
            if port.is_open():
                port.close()

    def append_out_port(self, port):
        if not isinstance(port, MidiPort):
            raise TypeError('Must be of type MidiPort.')
        if port not in self._out_ports:
            self._out_ports.append(port)

    def append_in_port(self, port):
        if not isinstance(port, MidiPort):
            raise TypeError('Must be of type MidiPort.')
        if port not in self._in_ports:
            self._in_ports.append(port)


class SystemMidiPortParser:
    """
    System-specific way of parsing MIDI port names.
    """

    def __init__(self, in_ports, out_ports):
        lp_in_ports = list(filter(lambda port: re.search(r'Mini\s*MK3',
                                                         port,
                                                         re.IGNORECASE),
                                  in_ports))
        lp_out_ports = list(filter(lambda port: re.search(r'Mini\s*MK3',
                                                          port,
                                                          re.IGNORECASE),
                                   out_ports))
        self._found_clients = []
        self._parse(lp_in_ports,
                    in_ports,
                    MidiPort.IN)
        self._parse(lp_out_ports,
                    out_ports,
                    MidiPort.OUT)

    def _parse(self, lp_ports, ports, direction):
        ClientData = namedtuple('ClientData',
                                ['client_name',
                                 'client_number',
                                 'ports'])
        PortData = namedtuple('PortData',
                              ['port_name',
                               'port_number',
                               'port_index',
                               'system_port_name',
                               'direction'])
        for index, system_port_name in enumerate(lp_ports):
            linux_match = re.match(r'(.+):(.+)\s(\d+):(\d+)$', system_port_name)  # noqa
            windows_match = re.match(r'^(.+)\s\d+$', system_port_name)
            windows_match2 = re.match(r'(.+)\s\((.+)\)\s\d+$', system_port_name)  # noqa
            mac_match = re.match(r'(.+\sMK3)\s(.+)$', system_port_name)

            client_name = ''
            port_name = ''
            client_number = int(index + 1)
            port_number = int(index + 1)

            if linux_match:
                client_name = linux_match.group(1)
                port_name = linux_match.group(2)
                client_number = int(linux_match.group(3))
                port_number = int(linux_match.group(4))
            elif windows_match and not windows_match2:
                client_name = windows_match.group(1)
                port_name = system_port_name
            elif windows_match2:
                client_name = windows_match2.group(2)
                port_name = system_port_name
            elif mac_match:
                client_name = mac_match.group(1)
                port_name = system_port_name

            port_index = ports.index(system_port_name)
            port_data = PortData(port_name,
                                 port_number,
                                 port_index,
                                 system_port_name,
                                 direction)
            existing_client_data = list(filter(lambda c: c.client_name == client_name,  # noqa
                                        self._found_clients))
            existing_client_data = (existing_client_data[0]
                                    if len(existing_client_data) > 0
                                    else None)
            if not existing_client_data:
                client_data = ClientData(client_name,
                                         client_number,
                                         [])
                client_data.ports.append(port_data)
                self._found_clients.append(client_data)
            else:
                existing_client_data.ports.append(port_data)

    @property
    def found_clients(self):
        return self._found_clients



"""
Scroll text from right to left across the Launchpad's surface.
"""

import sys
from lpminimk3 import Mode, find_launchpads
from lpminimk3.graphics import Text


def main():
    lp = find_launchpads()[0]  # Get the first available launchpad
    lp.open()  # Open device for reading and writing on MIDI interface (by default)  # noqa

    lp.mode = Mode.PROG  # Switch to the programmer mode

    print('Watch text scroll across the Launchpad\'s surface.\n'  # noqa
          'Press Ctrl+C to quit.\n')

    try:
        lp.grid.render(Text(' Hello, world!').scroll())
    except KeyboardInterrupt:
        return 1


if __name__ == '__main__':
    sys.exit(main())



"""
Match MIDI messages with incoming MIDI events.
"""

from abc import ABC
from .midi_messages import Lighting, Constants


class Match(ABC):
    """
    A set of rules for filtering MIDI events.
    """

    def contains(self, message):
        return False


class ButtonMatch(Match):
    """
    A set of rules for filtering button events.
    """

    def __init__(self, buttons, type):
        self._buttons = buttons
        self._type = type.lower().replace('|', '_')

    def contains(self, message):
        """
        Returns `True` if message `message` is a valid button
        message, otherwise returns `False`.
        """
        button_messages = self._determine_messages()
        for button_message in button_messages:
            if message == button_message:
                return True
        return False

    def _determine_messages(self):
        messages = []
        for button in self._buttons:
            header = (Constants.MidiWord.NOTE_HEADER
                      if button.parent == 'grid'
                      else Constants.MidiWord.CC_HEADER)
            if self._type == 'press':
                messages.append(Lighting(header, button.midi_value, Constants.MIDI_MAX_VALUE).data)  # noqa
            elif self._type == 'release':
                messages.append(Lighting(header, button.midi_value, Constants.MIDI_MIN_VALUE).data)  # noqa
            elif self._type == 'press_release':
                messages.append(Lighting(header, button.midi_value, Constants.MIDI_MAX_VALUE).data)  # noqa
                messages.append(Lighting(header, button.midi_value, Constants.MIDI_MIN_VALUE).data)  # noqa
        return messages



"""
Software representation of the Launchpad Mini MK3.
"""

from .__version__ import __version__, VERSION  # noqa
from ._core.components import Grid, Panel, Led, ButtonFace  # noqa
from ._core.utils import Interface, Mode, Layout, ButtonEvent  # noqa
from ._lpminimk3 import LaunchpadMiniMk3, find_launchpads  # noqa



"""
Random party mode with some interactivity

Author: Matthew Wachter
"""

from lpminimk3 import Mode, find_launchpads, colors
import random
import math


def main():  # noqa
    lp = find_launchpads()[0]  # Get the first available launchpad
    lp.open()  # Open device for reading and writing on MIDI interface (by default)  # noqa

    lp.mode = Mode.PROG  # Switch to the programmer mode

    for led in lp.panel.led_range():
        del led.color  # Turn off LED

    try:
        while True:
            order = [i for i in range(0, 64)]
            for i in range(0, len(order)):

                random.shuffle(order)
                led_n = order[i]
                led = lp.grid.led(led_n % 8, math.floor(led_n / 8))
                r = random.randint(-6, 6)
                if r > 4:
                    led.color = colors.ColorPalette.Red.SHADE_2
                elif r > 0:
                    led.color = colors.ColorPalette.Red.SHADE_9
                elif r == 0:
                    led.color = colors.ColorPalette.Yellow.SHADE_3
                else:
                    led.color = 0

                l_time = random.randint(1, 8)
                ev = lp.panel.buttons().poll_for_event(type='press',
                                                       timeout=l_time / 16)
                if ev and ev.button:
                    if r > 0:
                        ev.button.led.color = colors.ColorPalette.Green.SHADE_7
                    else:
                        ev.button.led.color = colors.ColorPalette.Green.SHADE_10  # noqa
    except KeyboardInterrupt:
        print('stopping')


if __name__ == '__main__':
    main()



"""
Render script for rendering texts, bitmaps and movies.
"""

import json
import sys
from argparse import ArgumentParser, FileType
from lpminimk3 import find_launchpads, Mode
from lpminimk3.graphics import Text, Bitmap, Movie


def render_character(character,
                     *,
                     on_screen=False,
                     lp=None):
    if on_screen:
        Text(character).print()
    elif lp:
        lp.grid.render(Text(character))
    else:
        raise RuntimeError("No Launchpad connected.")


def render_text(text,
                *,
                on_screen=False,
                count=-1,
                lp=None):
    if on_screen:
        Text(text).print()
    elif lp:
        lp.grid.render(Text(text).scroll(count=count))


def render_bitmap(filename, *, on_screen=False, lp=None):
    if on_screen:
        Bitmap(filename).print()
    elif lp:
        lp.grid.render(Bitmap(filename))


def render_movie(filename,
                 *,
                 on_screen=False,
                 count=-1,
                 lp=None):
    if on_screen:
        Movie(filename).print()
    elif lp:
        lp.grid.render(Movie(filename).play(count=count))


def find_lp():
    found_lps = find_launchpads()
    if len(found_lps):
        lp = found_lps[0]
        lp.open()
        lp.mode = Mode.PROG
        return lp
    return None


def _render(args, *, lp=None):
    if args.f:
        data = json.load(args.f)
        if "bitmap" in data:
            render_bitmap(args.f.name,
                          on_screen=args.s,
                          lp=lp)
        elif "frames" in data:
            render_movie(args.f.name,
                         on_screen=args.s,
                         count=args.c,
                         lp=lp)
    elif len(args.text) == 1:
        render_character(args.text,
                         on_screen=args.s,
                         lp=lp)
    else:
        render_text(args.text,
                    on_screen=args.s,
                    count=args.c,
                    lp=lp)


def main(args=None):
    args = args if args else sys.argv[1:]
    try:
        parser = ArgumentParser(description="Render texts, bitmaps, "
                                            "and movies on the "
                                            "Launchpad Mini MK3")

        parser.add_argument("-c",
                            type=int,
                            metavar="COUNT",
                            help="Stop after looping COUNT times")
        parser.add_argument("-f",
                            type=FileType('r', encoding='latin-1'),
                            metavar="FILE",
                            help="Bitmap/movie file to render")
        parser.add_argument("-s",
                            action="store_true",
                            help="Print to screen")
        parser.add_argument("text",
                            nargs="?",
                            metavar="TEXT",
                            help="Text to render")
        parser.add_argument("--version",
                            action="version",
                            version="%(prog) 0.1")
        if len(args):
            args = parser.parse_args(args)
            lp = find_lp()
            _render(args, lp=lp)
        else:
            parser.print_help()
            return 1
    except KeyboardInterrupt:
        return 1


if __name__ == '__main__':
    sys.exit(main())



import os
import json
import jsonschema
from ..midi_messages import Constants


class GlyphDictionary:
    def __init__(self,
                 json_filename, *,
                 schema_filename='./schema/glyph.schema.json'):
        self._filename = self._determine_abspath(json_filename)
        self._data = self._load(json_filename)
        schema = self._load(schema_filename)
        self._validate(self._data, schema)

    def __iter__(self):
        for glyph, bitmap_data in self._data.items():
            yield glyph, bitmap_data

    def __contains__(self, unicode):
        return unicode in self._data['glyphs']

    def __getitem__(self, unicode):
        return self._data['glyphs'][unicode]

    def __repr__(self):
        return f"GlyphDictionary(filename='{os.path.basename(self.filename)}')"

    def __str__(self):
        return (str(self._data['glyphs'])
                if self._data
                else '')

    @property
    def filename(self):
        return self._filename

    @property
    def data(self):
        return self._data

    def _load(self, filename):
        data = None
        filename = self._determine_abspath(filename)
        with open(filename) as f:
            data = json.load(f)
        return data

    def _validate(self, data, schema):
        try:
            jsonschema.validate(instance=data, schema=schema)
        except jsonschema.exceptions.ValidationError:
            raise ValueError('Invalid JSON file format.')

    def _determine_abspath(self, filename):
        if not os.path.isabs(filename):
            current_dir = os.path.dirname(__file__)
            return os.path.join(current_dir, filename)
        return filename


class BitmapDocument:
    def __init__(self,
                 json_filename, *,
                 schema_filename='./schema/bitmap.schema.json'):
        assert json_filename
        self._filename = self._determine_abspath(json_filename)
        self._data = self._load(json_filename)
        schema = self._load(schema_filename)
        self._validate(self._data, schema)

    def __repr__(self):
        return f"BitmapDocument(filename='{self.filename}')"

    def __str__(self):
        return repr(self)

    @property
    def filename(self):
        return self._filename

    @property
    def bitmap_data(self):
        return self._data['bitmap']['data']

    @property
    def bitmap_config(self):
        return self._data['bitmap']['config']

    @property
    def data(self):
        return self._data

    def _load(self, filename):
        data = None
        filename = self._determine_abspath(filename)
        with open(filename) as f:
            data = json.load(f)
        return data

    def _validate(self, data, schema):
        try:
            jsonschema.validate(instance=data, schema=schema)
        except jsonschema.exceptions.ValidationError:
            raise ValueError('Invalid JSON file format.')

    def _determine_abspath(self, filename):
        if not os.path.isabs(filename):
            current_dir = os.path.dirname(__file__)
            return os.path.join(current_dir, filename)
        return filename


class MovieDocument:
    def __init__(self,
                 json_filename, *,
                 schema_filename='./schema/movie.schema.json'):
        assert json_filename
        self._filename = self._determine_abspath(json_filename)
        self._data = self._load(json_filename)
        schema = self._load(schema_filename)
        self._validate(self._data, schema)

    def __repr__(self):
        return f"MovieDocument(filename='{self.filename}')"

    def __str__(self):
        return repr(self)

    @property
    def filename(self):
        return self._filename

    @property
    def framerate(self):
        return self._data['framerate']

    @property
    def frames(self):
        return self._data['frames']

    @property
    def data(self):
        return self._data

    def _load(self, filename):
        data = None
        filename = self._determine_abspath(filename)
        with open(filename) as f:
            data = json.load(f)
        return data

    def _validate(self, data, schema):
        try:
            jsonschema.validate(instance=data, schema=schema)
        except jsonschema.exceptions.ValidationError:
            raise ValueError('Invalid JSON file format.')

    def _determine_abspath(self, filename):
        if not os.path.isabs(filename):
            current_dir = os.path.dirname(__file__)
            return os.path.join(current_dir, filename)
        return filename


class LightingType:
    FLASH = 'flash'
    PULSE = 'pulse'
    RGB = 'rgb'

    def __init__(self, type_str):
        self._type_str = type_str

    @property
    def midi_value(self):
        if self._type_str == LightingType.FLASH:
            return Constants.LightingType.FLASH
        elif self._type_str == LightingType.PULSE:
            return Constants.LightingType.PULSE
        elif self._type_str == LightingType.RGB:
            return Constants.LightingType.RGB
        return Constants.LightingType.STATIC


class LightingColor:
    def __init__(self, *, id=0, rgb=[0, 0, 0]):
        self._id = id
        self._rgb = rgb

    def __repr__(self):
        return ('LightingColor('
                f'id={self.id}, '
                f'rgb={self.rgb})')

    def __str__(self):
        return repr(self)

    @property
    def id(self):
        return self._id

    @property
    def rgb(self):
        return self._rgb


class LightingState:
    def __init__(self, state=None):
        self._color = (LightingColor(**state['color'])
                       if state
                       else LightingColor())

    def __repr__(self):
        return ('LightingState('
                f'color_id={self._color.id}, '
                f'rgb={self._color.rgb})')

    def __str__(self):
        return repr(self)

    @property
    def color(self):
        return self._color


class LightingData:
    def __init__(self,
                 lighting_type, *,
                 on_state=None,
                 off_state=None):
        self._lighting_type = lighting_type
        self._on_state = (LightingState(on_state)
                          if on_state
                          else None)
        self._off_state = (LightingState(off_state)
                           if off_state
                           else None)

    def __repr__(self):
        return ("LightingData("
                f"on_state={self.on_state}, "
                f"off_state={self.off_state})")

    def __str__(self):
        return repr(self)

    @property
    def on_state(self):
        return self._on_state

    @property
    def off_state(self):
        return self._off_state


class BitConfig:
    def __init__(self, name=None, config_data=None):
        self._name = name
        self._data = config_data

    def __repr__(self):
        return f"BitConfig(name='{self._name}')"

    def __str__(self):
        return repr(self)

    @property
    def lighting_type(self):
        return (LightingType(self._data['lighting_type']).midi_value
                if self._data and 'lighting_type' in self._data
                else Constants.LightingType.STATIC)

    @property
    def name(self):
        return self._name

    @property
    def lighting_data(self):
        if self._data:
            return LightingData(self.lighting_type,
                                **self._data.get('lighting_data'))
        return None


class BitmapConfig:
    def __init__(self, config_data):
        self._data = self._load(config_data)

    def __getitem__(self, name):
        if self._data and name in self._data:
            return self._data[name]
        return BitConfig()

    def _load(self, config_data):
        config_dict = {}
        if config_data:
            for button_name, data in config_data.items():
                config_dict[button_name] = BitConfig(button_name, data)
        return config_dict

    def __repr__(self):
        return f"BitmapConfig('{str(self._data)}')"

    def __str__(self):
        return repr(self)



import os
import time
from abc import ABC
from functools import reduce
from ..colors._colors import ColorShade, ColorShadeStore, RgbColor
from ._parser import GlyphDictionary,\
                     BitmapDocument,\
                     MovieDocument
from ._renderer import CharacterRenderer,\
                       BitmapRenderer,\
                       MovieRenderer,\
                       FrameRenderer,\
                       TextRenderer
from ._utils import RawBitmap,\
                    RawBitmapMatrix,\
                    ScrollDirection,\
                    FlipAxis,\
                    Offset


class Renderable(ABC):
    @property
    def bits(self):
        pass

    @property
    def word_count(self):
        return 0

    def render(self, matrix):
        pass


class RenderableColor:
    def __init__(self, renderable, value):
        self._renderable = renderable
        self._value = value
        self._validate(value)

    def __repr__(self):
        return f'RenderableColor({self._value})'

    @property
    def color_id(self):
        return self._color_id

    def set(self, value):
        self._validate(value)
        return self._renderable

    def _validate(self, value):
        if not value:
            self._color_id = 0
        elif (not isinstance(value, ColorShade)
                and not isinstance(value, str)
                and not isinstance(value, int)
                and not isinstance(value, (tuple, list))):
            raise TypeError('Must be of type ColorShade '
                            'or str or int or tuple or list.')
        elif ((isinstance(value, str)
                and not ColorShadeStore().contains(value)
                and not RgbColor.is_valid(value))
                or (isinstance(value, (tuple, list))
                    and not RgbColor.is_valid(value))):
            raise ValueError('Invalid color.')
        elif RgbColor.is_valid(value):
            self._rgb_color = RgbColor(value)
        else:
            self._color_id = self._determine_color_id(value)

    def _determine_color_id(self, value):
        color_id = (value
                    if ColorShade.is_valid_id(value)
                    else -1)
        if color_id < 0:
            color_shade = ColorShadeStore().find(value)
            color_id = (color_shade.color_id
                        if color_shade
                        else color_id)
        if color_id < 0:
            raise ValueError(f'Color ID values must be between '
                             f'{ColorShade.MIN_COLOR_ID} and '
                             f'{ColorShade.MAX_COLOR_ID}.')
        return color_id


class CharacterTransform:
    def __init__(self, character, bitmap_data):
        self._character = character
        self._character_raw_bitmap = RawBitmap(bitmap_data)

    def __repr__(self):
        return ("CharacterTransform("
                f"character='{self._character}')")

    def shift_left(self, *, carry=None, count=1, circular=False):
        count = max(0, count)
        count = count % self._character.word_count
        character = self._character
        for _ in range(count):
            transformed_bitmap_data = []
            temp_bitmap_data = []
            new_carry = 0
            msb_address = self._character.word_count - 1
            for index, word in enumerate(character.raw_bitmap.data):
                if carry:
                    temp_bitmap_data.append((word >> 1) | ((carry & (1 << index))  # noqa
                                                            << (msb_address - index)))  # noqa
                else:
                    temp_bitmap_data.append(word >> 1)
                new_carry |= (word & 1) << index
            transformed_bitmap_data = temp_bitmap_data
            if circular and new_carry:
                temp_bitmap_data = []
                for index, word in enumerate(transformed_bitmap_data):  # noqa
                    temp_bitmap_data.append(word | (((new_carry & (1 << index))
                                                     >> index << msb_address)))
                transformed_bitmap_data = temp_bitmap_data
            character = Character(self._character.glyph,
                                  self._character_raw_bitmap.data,
                                  self._character.fg_color,
                                  self._character.bg_color,
                                  carry=new_carry,
                                  transformed_bitmap_data=transformed_bitmap_data,  # noqa
                                  offset=(self._character.offset.x + count,
                                          self._character.offset.y))
        return character

    def shift_right(self, *, carry=None, count=1, circular=False):
        count = max(0, count)
        count = count % self._character.word_count
        character = self._character
        for _ in range(count):
            transformed_bitmap_data = []
            temp_bitmap_data = []
            new_carry = 0
            msb_address = self._character.word_count - 1
            for index, word in enumerate(character.raw_bitmap.data):
                if carry:
                    temp_bitmap_data.append((word << 1) | ((carry & (1 << (msb_address - index)))  # noqa
                                                            >> (msb_address - index)))  # noqa
                else:
                    temp_bitmap_data.append(word << 1)
                new_carry |= (word & (1 << msb_address)) >> index
            transformed_bitmap_data = temp_bitmap_data
            if circular and new_carry:
                temp_bitmap_data = []
                for index, word in enumerate(transformed_bitmap_data):
                    temp_bitmap_data.append(word | ((new_carry & (1 << index))  # noqa
                                                    >> index))
                transformed_bitmap_data = temp_bitmap_data
            character = Character(self._character.glyph,
                                  self._character_raw_bitmap.data,
                                  self._character.fg_color,
                                  self._character.bg_color,
                                  carry=new_carry,
                                  transformed_bitmap_data=transformed_bitmap_data,  # noqa
                                  offset=(self._character.offset.x - count,
                                          self._character.offset.y))
        return character


class TextScroll:
    def __init__(self,
                 text,
                 period,
                 direction,
                 *,
                 cycle_func,
                 timeout,
                 count):
        if timeout:
            assert period < timeout
        if cycle_func:
            assert callable(cycle_func)

        self._text = text
        self._period = period
        self._direction = direction
        self._cycle_func = cycle_func
        self._timeout = -1 if not timeout else timeout
        self._count = -1 if not count else count

    def render(self, string, matrix):
        TextRenderer(self._text,
                     string,
                     self._period,
                     self._direction,
                     matrix,
                     timeout=self._timeout,
                     count=self._count,
                     cycle_func=self._cycle_func).render()

    def print(self, *,
              string,
              one,
              zero):
        time_left = self._timeout
        rotations_left = self._count
        while time_left and rotations_left:
            for _ in range(len(self._text) * string.character_to_render.word_count):  # noqa
                if self._direction == ScrollDirection.RIGHT:
                    string.shift_right()
                else:
                    string.shift_left()
                for index, bit in enumerate(string.character_to_render.raw_bitmap,  # noqa
                                            start=1):
                    if bit:
                        print(one, end='')
                    else:
                        print(zero, end='')
                    if index % string.character_to_render.word_count == 0:
                        print('\n', end='')
                time.sleep(self._period)
                time_left = (max(time_left - self._period, 0)
                             if time_left != -1
                             else time_left)
                os.system('clear')
                print('\r', end='')
            rotations_left = (max(rotations_left - 1, 0)
                              if rotations_left != -1
                              else rotations_left)


class MovieRoll:
    def __init__(self,
                 movie,
                 framerate,
                 *,
                 cycle_func,
                 count):
        if cycle_func:
            assert callable(cycle_func)

        self._movie = movie
        self._framerate = framerate
        self._cycle_func = cycle_func
        self._count = -1 if not count else count

    def render(self, movie, matrix):
        MovieRenderer(movie.raw_bitmaps,
                      self._framerate,
                      matrix,
                      count=self._count).render()

    def print(self, *,
              movie,
              one,
              zero):
        rotations_left = self._count
        period = 1 / self._framerate
        while rotations_left:
            for index, bit in enumerate(movie.frame_to_render.raw_bitmap,  # noqa
                                        start=1):
                if bit:
                    print(one, end='')
                else:
                    print(zero, end='')
                if index % movie.frame_to_render.word_count == 0:
                    print('\n', end='')

            movie.position = (0
                              if rotations_left
                              and (movie.position == len(movie.frames) - 1)
                              else movie.position)
            time.sleep(period)
            movie.skip()
            rotations_left = (max(rotations_left - 1, 0)
                              if rotations_left != -1
                              else rotations_left)
            os.system('clear')
            print('\r', end='')


class Character(Renderable):
    def __init__(self,
                 glyph,
                 bitmap_data,
                 fg_color,
                 bg_color,
                 *,
                 carry=None,
                 transformed_bitmap_data=None,
                 offset=None):
        assert isinstance(fg_color, RenderableColor)
        assert isinstance(bg_color, RenderableColor)

        self._glyph = glyph
        self._raw_bitmap = RawBitmap(bitmap_data)
        self._fg_color = fg_color
        self._bg_color = bg_color
        self._carry = carry
        self._raw_bitmap_transformed = (RawBitmap(transformed_bitmap_data)
                                        if transformed_bitmap_data
                                        else None)
        self._offset = (Offset(*offset)
                        if offset
                        else Offset())
        self._angle = 0
        self._flip_axis = ''

    def __repr__(self):
        return ("Character("
                f"glyph='{self._glyph}', "
                f"offset={self._offset}, "
                f"carry={self.carry})")

    def __str__(self):
        return f'{self._glyph}'

    def __iter__(self):
        return self.bits

    @Renderable.bits.getter
    def bits(self):
        for bit in self._raw_bitmap:
            yield bit

    @Renderable.word_count.getter
    def word_count(self):
        return self._raw_bitmap.word_count

    @property
    def glyph(self):
        return self._glyph

    @property
    def fg_color(self):
        return self._fg_color

    @fg_color.setter
    def fg_color(self, color):
        assert isinstance(color, RenderableColor)
        self._fg_color = color

    @property
    def bg_color(self):
        return self._bg_color

    @bg_color.setter
    def bg_color(self, color):
        assert isinstance(color, RenderableColor)
        self._bg_color = color

    @property
    def carry(self):
        return self._carry

    @property
    def raw_bitmap(self):
        return (self._raw_bitmap_transformed
                if self._raw_bitmap_transformed
                else self._raw_bitmap)

    @property
    def offset(self):
        return self._offset

    def render(self, matrix):
        CharacterRenderer(self,
                          matrix,
                          angle=self._angle,
                          flip_axis=self._flip_axis).render()

    def print(self):
        print(self.raw_bitmap)

    def shift_left(self, *, carry=None, count=1, circular=False):
        return CharacterTransform(self,
                                  self._raw_bitmap.data)\
                                          .shift_left(carry=carry,
                                                      count=count,
                                                      circular=circular)

    def shift_right(self, *, carry=None, count=1, circular=False):
        return CharacterTransform(self,
                                  self._raw_bitmap.data)\
                                          .shift_right(carry=carry,
                                                       count=count,
                                                       circular=circular)

    def rotate(self, angle):
        self._angle = angle


class String(Renderable):
    def __init__(self, text, *, fg_color, bg_color):
        assert isinstance(fg_color, RenderableColor)
        assert isinstance(bg_color, RenderableColor)

        if text and not isinstance(text, str):
            raise TypeError("text must be of type 'str'.")

        glyph_dicts = self._create_glyph_dicts()
        characters = self._create_characters(text,
                                             glyph_dicts,
                                             fg_color,
                                             bg_color)
        self._text = text
        self._characters = list(characters)
        self._angle = 0
        self._text_scroll = None
        self._flip_axis = ''

    def __repr__(self):
        return (f"String('{self}')")

    def __str__(self):
        char_list = list(reduce(lambda literal, char: literal + str(char),
                                self._characters, ''))
        return f"{''.join(char_list)}"

    @Renderable.bits.getter
    def bits(self):
        if self.angle:
            matrix = RawBitmapMatrix(self.character_to_render.raw_bitmap)
            for bit in matrix.rotated_range(self.angle,
                                            flip_axis=self.flip_axis):
                yield bit
        elif self.flip_axis:
            matrix = RawBitmapMatrix(self.character_to_render.raw_bitmap)
            for bit in matrix.flipped_range(self._flip_axis):
                yield bit
        else:
            for bit in self.character_to_render.raw_bitmap:
                yield bit

    @Renderable.word_count.getter
    def word_count(self):
        return self.character_to_render.word_count

    @property
    def fg_color(self):
        return self.character_to_render.fg_color

    @fg_color.setter
    def fg_color(self, color):
        assert isinstance(color, RenderableColor)
        self.character_to_render.fg_color = color

    @property
    def bg_color(self):
        return self.character_to_render.bg_color

    @bg_color.setter
    def bg_color(self, color):
        assert isinstance(color, RenderableColor)
        self.character_to_render.bg_color = color

    @property
    def characters(self):
        return self._characters

    @property
    def character_to_render(self):
        return self._characters[0]

    @property
    def angle(self):
        return self._angle

    @property
    def flip_axis(self):
        return self._flip_axis

    @flip_axis.setter
    def flip_axis(self, flip_axis):
        assert isinstance(flip_axis, str)
        assert flip_axis in FlipAxis.XY
        self._flip_axis = flip_axis

    @property
    def text_scroll(self):
        return self._text_scroll

    @text_scroll.setter
    def text_scroll(self, text_scroll):
        if not isinstance(text_scroll, TextScroll):
            raise ValueError("Must be of type 'TextScroll'.")
        self._text_scroll = text_scroll

    def render(self, matrix):
        if self._text_scroll:
            self._text_scroll.render(self, matrix)
        else:
            CharacterRenderer(self.character_to_render,
                              matrix,
                              angle=self.angle,
                              flip_axis=self.flip_axis).render()

    def print(self, *,
              one='X',
              zero=' '):
        if self._text_scroll:
            self._text_scroll.print(string=self,
                                    one=one,
                                    zero=zero)
        else:
            self._print_in_console(one=one, zero=zero)

    def shift_left(self, *, count=1, circular=True):
        count = 0 if not isinstance(count, int) or count < 0 else count
        if len(self._characters) == 1:
            self._characters[0] = self._characters[0].shift_left(count=count,
                                                                 circular=circular)  # noqa
        else:
            for _ in range(count):
                shifted_characters = []
                carry = 0
                if circular:
                    shifted_first_character = self._characters[0].shift_left()
                    carry = shifted_first_character.carry
                for character in reversed(self._characters):
                    new_character = character.shift_left(carry=carry)
                    carry = new_character.carry
                    shifted_characters.append(new_character)
                shifted_characters.reverse()
                self._characters = shifted_characters

    def shift_right(self, *, count=1, circular=True):
        count = 0 if not isinstance(count, int) or count < 0 else count
        if len(self._characters) == 1:
            self._characters[0] = self._characters[0].shift_right(count=count,
                                                                  circular=circular)  # noqa
        else:
            for _ in range(count):
                shifted_characters = []
                carry = 0
                if circular:
                    shifted_last_character = self._characters[-1].shift_right()
                    carry = shifted_last_character.carry
                for character in self._characters:
                    new_character = character.shift_right(carry=carry)
                    carry = new_character.carry
                    shifted_characters.append(new_character)
                self._characters = shifted_characters

    def rotate(self, angle):
        self._angle = angle

    def _create_characters(self, text, dicts, fg_color, bg_color):
        characters = []
        text = '\u0000' if not text else text
        for glyph in text:
            for glyph_dict in dicts:
                if glyph in glyph_dict:
                    characters.append(Character(glyph,
                                                glyph_dict[glyph],
                                                fg_color,
                                                bg_color))
                    break
        return characters

    def _create_glyph_dicts(self):
        glyph_dicts = []
        current_dir = os.path.dirname(__file__)
        glyph_dict_dir = os.path.join(current_dir, 'glyphs')
        for filename in os.listdir(glyph_dict_dir):
            if filename.endswith('.glyph.json'):
                glyph_dicts.append(GlyphDictionary(os.path.join(glyph_dict_dir, filename)))  # noqa
        return glyph_dicts

    def _print_in_console(self, one='X', zero=' '):
        matrix = RawBitmapMatrix(self.character_to_render.raw_bitmap)
        if self._flip_axis and not self._angle:
            for index, bit in enumerate(matrix.flipped_range(self._flip_axis),   # noqa
                                        start=1):  # noqa
                self._print_bit(bit, index, one=one, zero=zero)
        elif self._angle:
            for index, bit in enumerate(matrix.rotated_range(self._angle,
                                                             flip_axis=self._flip_axis),  # noqa
                                        start=1):
                self._print_bit(bit, index, one=one, zero=zero)
        else:
            for index, bit in enumerate(self.character_to_render.raw_bitmap,
                                        start=1):
                self._print_bit(bit, index, one=one, zero=zero)

    def _print_bit(self, bit, index, *, one, zero):
        if bit:
            print(one[0], end='')
        elif 'LOGLEVEL' in os.environ:
            print('.', end='')
        else:
            print(zero[0], end='')
        if index % self.character_to_render.word_count == 0:
            print('\n', end='')


class Bitmap(Renderable):
    def __init__(self,
                 filename, *,
                 fg_color,
                 bg_color):
        assert isinstance(filename, str)
        assert isinstance(fg_color, RenderableColor)
        assert isinstance(bg_color, RenderableColor)
        self._bitmap_document = BitmapDocument(filename)
        self._raw_bitmap = RawBitmap(self._bitmap_document.bitmap_data,
                                     self._bitmap_document.bitmap_config)
        self._fg_color = fg_color
        self._bg_color = bg_color

    def __repr__(self):
        return (f"Bitmap('{self}')")

    def __str__(self):
        return repr(self)

    @Renderable.bits.getter
    def bits(self):
        for bit in self.raw_bitmap:
            yield bit

    @Renderable.word_count.getter
    def word_count(self):
        return self._raw_bitmap.word_count

    @property
    def filename(self):
        return self._bitmap_document.filename

    @property
    def raw_bitmap(self):
        return self._raw_bitmap

    @property
    def fg_color(self):
        return self._fg_color

    @fg_color.setter
    def fg_color(self, color):
        assert isinstance(color, RenderableColor)
        self._fg_color = color

    @property
    def bg_color(self):
        return self._bg_color

    @bg_color.setter
    def bg_color(self, color):
        assert isinstance(color, RenderableColor)
        self._bg_color = color

    def render(self, matrix):
        BitmapRenderer(self.raw_bitmap,
                       matrix,
                       fg_color=self.fg_color,
                       bg_color=self.bg_color).render()

    def print(self, *,
              one='X',
              zero=' '):
        for index, bit in enumerate(self.raw_bitmap,
                                    start=1):
            self._print_bit(bit, index, one=one, zero=zero)

    def _print_bit(self, bit, index, *, one, zero):
        if bit:
            print(one[0], end='')
        elif 'LOGLEVEL' in os.environ:
            print('.', end='')
        else:
            print(zero[0], end='')
        if index % self.raw_bitmap.word_count == 0:
            print('\n', end='')


class Movie(Renderable):
    def __init__(self,
                 filename, *,
                 fg_color,
                 bg_color):
        assert isinstance(filename, str)
        assert isinstance(fg_color, RenderableColor)
        assert isinstance(bg_color, RenderableColor)
        self._movie_document = MovieDocument(filename)
        self._raw_bitmaps = list(map(lambda frame: RawBitmap(
                                              frame["data"],
                                              frame["config"]),
                                 self._movie_document.frames))
        self._fg_color = fg_color
        self._bg_color = bg_color
        self._position = 0 if len(self._raw_bitmaps) > 0 else -1
        self._framerate = self._movie_document.framerate
        self._roll = None
        self._angle = 0
        self._flip_axis = 'x'

    def __repr__(self):
        return (f"Movie('{self}')")

    def __str__(self):
        return repr(self)

    @Renderable.bits.getter
    def bits(self):
        for bit in self.frame_to_render.raw_bitmap:
            yield bit

    @Renderable.word_count.getter
    def word_count(self):
        return self.frame_to_render.raw_bitmap.word_count

    @property
    def fg_color(self):
        return self._fg_color

    @fg_color.setter
    def fg_color(self, color):
        assert isinstance(color, RenderableColor)
        self._fg_color = color

    @property
    def bg_color(self):
        return self._bg_color

    @bg_color.setter
    def bg_color(self, color):
        assert isinstance(color, RenderableColor)
        self._bg_color = color

    @property
    def filename(self):
        return self._movie_document.filename

    @property
    def framerate(self):
        return self._framerate

    @framerate.setter
    def framerate(self, framerate):
        self._framerate = framerate

    @property
    def raw_bitmaps(self):
        return self._raw_bitmaps

    @property
    def frames(self):
        return list(map(lambda bitmap: Frame(bitmap), self._raw_bitmaps))

    @property
    def position(self):
        return self._position

    @position.setter
    def position(self, position):
        assert isinstance(position, int)
        self._position = position

    @property
    def frame_to_render(self):
        return (self.frames[self._position]
                if len(self.frames)
                else None)

    @property
    def first_frame(self):
        return (self.frames[0]
                if len(self.frames)
                else None)

    @property
    def last_frame(self):
        return (self.frames[-1]
                if len(self.frames)
                else None)

    @property
    def roll(self):
        return self._roll

    @roll.setter
    def roll(self, roll):
        if not isinstance(roll, MovieRoll):
            raise ValueError("Must be of type 'MovieRoll'.")
        self._roll = roll

    def render(self, matrix):
        if self._roll:
            self._roll.render(self, matrix)
        else:
            FrameRenderer(self.frame_to_render.raw_bitmap,
                          matrix).render()

    def skip(self, frame_count=1, *, ref=None):
        assert isinstance(frame_count, int)
        if not frame_count:
            return ref
        forward = frame_count > 0
        new_position = frame_count + self.position
        if forward:
            if new_position < len(self.frames):
                self._position = new_position
        else:
            if new_position >= 0:
                self._position = new_position

        return ref

    def print(self, *,
              one='X',
              zero=' '):
        if self._roll:
            self._roll.print(movie=self,
                             one=one,
                             zero=zero)
        else:
            self._print_in_console(one=one, zero=zero)

    def _print_bit(self, bit, index, *, one, zero):
        if bit:
            print(one[0], end='')
        elif 'LOGLEVEL' in os.environ:
            print('.', end='')
        else:
            print(zero[0], end='')
        if index % self.frame_to_render.word_count == 0:
            print('\n', end='')

    def _print_in_console(self, one='X', zero=' '):
        for index, bit in enumerate(self.frame_to_render.raw_bitmap,
                                    start=1):
            self._print_bit(bit, index, one=one, zero=zero)


class Frame(Renderable):
    def __init__(self, data):
        assert (isinstance(data, dict)
                or isinstance(data, RawBitmap))
        self._raw_bitmap = (RawBitmap(data['data'],
                                      data['config'])
                            if isinstance(data, dict)
                            else data)

    def __repr__(self):
        return (f"Frame('{repr(self._raw_bitmap.data)}')")

    def __str__(self):
        return repr(self)

    @Renderable.bits.getter
    def bits(self):
        for bit in self._raw_bitmap:
            yield bit

    @Renderable.word_count.getter
    def word_count(self):
        return self._raw_bitmap.word_count

    @property
    def raw_bitmap(self):
        return self._raw_bitmap

    def render(self, matrix):
        FrameRenderer(self._raw_bitmap,
                      matrix).render()

    def print(self, *,
              one='X',
              zero=' '):
        for index, bit in enumerate(self.raw_bitmap,
                                    start=1):
            self._print_bit(bit, index, one=one, zero=zero)

    def _print_bit(self, bit, index, *, one, zero):
        if bit:
            print(one[0], end='')
        elif 'LOGLEVEL' in os.environ:
            print('.', end='')
        else:
            print(zero[0], end='')
        if index % self.raw_bitmap.word_count == 0:
            print('\n', end='')



"""
Scroll text from left to right along the Launchpad's surface,
lighting a specified region.
"""

from lpminimk3 import Mode, find_launchpads
from lpminimk3.graphics import Text
from lpminimk3.region import Labeled as LabeledRegion
import random


def cycle_func(fraction, lp):
    labeled_region = LabeledRegion()  # Region of labeled buttons
    scroll_index = int(fraction * len(labeled_region))  # position in scroll
    for index, led in enumerate(lp.panel.led_range(region=labeled_region)):  # Loop through all labeled buttons  # noqa
        if scroll_index == index:
            led.color = 'green4'  # Light up LED to green
        else:
            del led.color  # Turn off all other LEDs


def main():
    lp = find_launchpads()[0]  # Get the first available launchpad
    lp.open()  # Open device for reading and writing on MIDI interface (by default)  # noqa

    lp.mode = Mode.PROG  # Switch to the programmer mode

    print('Watch text scroll across the Launchpad\'s surface.\n'  # noqa
          'Press Ctrl+C to quit.\n')

    while True:
        try:
            lp.grid.render(Text(' The entrance of your word giveth light...')
                           .scroll(count=1,
                                   cycle_func=cycle_func)
                           .rotate(-90)
                           .fg_color.set(random.randint(1, 127)))
        except KeyboardInterrupt:
            print('\n')
            break


if __name__ == '__main__':
    main()



import sys
import asyncio
import json
from collections import namedtuple
from argparse import ArgumentParser
from websockets import connect
from lpminimk3 import find_launchpads, Mode
from lpminimk3.graphics import Frame


def find_lps():
    lps = []
    for lp in find_launchpads():
        lp.open()
        lp.mode = Mode.PROG
        lps.append(lp)
    return lps


async def sync_with_server(lps, host, port):
    uri = f"ws://{host}:{port}/sync"
    async with connect(uri) as websocket:
        while True:
            data = await websocket.recv()
            data = json.loads(data)
            for lp in lps:
                lp.grid.render(Frame(data))


def init_parser(port):
    parser = ArgumentParser(description="Sync client example "
                                        "for lpminimk3")
    parser.add_argument("-i",
                        "--ip",
                        required=True,
                        help="Server IP")
    parser.add_argument("-p",
                        "--port",
                        type=int,
                        help=f"Server port (Default: {port})")
    return parser


async def main(*, ip=None, port=7654):
    Args = namedtuple("Args", ["ip", "port"])
    args = Args(ip, port)
    parser = init_parser(port)
    try:
        if len(sys.argv) > 1 or not ip:
            args = parser.parse_args(sys.argv[1:])
        lps = find_lps()
        ip = args.ip if args.ip else ip
        port = args.port if args.port else port
        if not ip:
            print("No IP provided.", file=sys.stderr)
            parser.print_help()
            return 1
        else:
            await sync_with_server(lps, ip, port)
    except KeyboardInterrupt:
        return 1


if __name__ == "__main__":
    asyncio.run(main())
